/****************************************************************************/
/*                  Copyright 2000, Trustees of Boston University.          */
/*                               All Rights Reserved.                       */
/*                                                                          */
/* Permission to use, copy, or modify this software and its documentation   */
/* for educational and research purposes only and without fee is hereby     */
/* granted, provided that this copyright notice appear on all copies and    */
/* supporting documentation.  For any other uses of this software, in       */
/* original or modified form, including but not limited to distribution in  */
/* whole or in part, specific prior permission must be obtained from Boston */
/* University.  These programs shall not be used, rewritten, or adapted as  */
/* the basis of a commercial software or hardware product without first     */
/* obtaining appropriate licenses from Boston University.  Boston University*/
/* and the author(s) make no representations about the suitability of this  */
/* software for any purpose.  It is provided "as is" without express or     */
/* implied warranty.                                                        */
/*                                                                          */
/****************************************************************************/
/*                                                                          */
/* Utility Name: Evaluate tcpdump log for time line, counts, critical path  */
/* Authors:      Paul Barford                                               */
/* First Rev:    v1.0  2/28/99                                              */
/*               v1.1  12/8/99                                              */
/* Stable:       v1.2   2/24/00                                             */
/*                                                                          */
/* Description:                                                             */
/*                                                                          */
/* This program parses two tcpdump log files taken at end points of an end  */
/* to end connection and generates a number of output files.  The first is  */
/* a file which can be used in xgraph to display the time line for each tcp */
/* connection that is tracked in the log.  The second is a file which has   */
/* the count of packets transferred at each end point for specified time    */
/* bins.  The third is a file with the number of bytes transferred in each  */
/* of the time bins.  The program also prints the number of SYN, FIN and    */
/* ACK packets that appear in the trace at each end point.                  */
/*                                                                          */
/* Inputs:                                                                  */
/*                                                                          */
/* The parameters for the program are listed in the DEFINE area of the      */
/* program header.  This program reads in all of the data from the log file */
/* The tcpdump log files should be generated using the following command:   */
/*                                                                          */
/*                  tcpdummp -vv port {http port} and host {hostname}       */
/*                                                                          */
/* Time stamps in tcpdump are of the format hh:mm:ss.frac.  This program    */
/* expects two tcpdump trace files to be specified on the command line and  */
/* should be started up as follows:                                         */
/*                                                                          */
/*                  tcpeval  remote_trace server_trace                      */
/*                                                                          */
/*  NOTE the order of the input files (remote, server) is not important     */
/*                                                                          */
/* Outputs:                                                                 */
/*                                                                          */
/* The program outputs three files:  the first which is formatted           */
/* to display time line graphs specifically in xgraph.  The second file     */
/* contains a single column which is the counts of packets which occured    */
/* in each time interval.  The file is called tcpcounts.txt.  The third     */
/* file contains the number of bytes transfered during the same time        */
/* interval used for the counts.  This file is called tcpbytes.txt.  It     */
/* This program also prints the total number of packets of each type - SYN, */
/* FIN, ACK, data and dropped packets that were transmitted.                */
/*                                                                          */
/****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>

#define MAXTRACE 500000      /* MAX number of packets in input file         */
#define MAXSEQ 100000        /* MAX number of packets per downloaded file   */
#define MAXPKTS 10           /* MAX number of packets between acks          */
#define USINTRVL 100000      /* Interval for counting in microseconds       */
#define SECINTRVL 0          /* Interval for counting in seconds            */
#define DEBUG 0              /* set DEBUG to 1 to get print statements       */
#define GRAPHTOOL 0          /* if GRAPHTOOL = 0 use gnuplot, 1 use xgraph  */

/* Since port addresses can wrap around during a test we must still be able */
/* to separate transfers.  If this happens, we sort by both port numbers    */
/* and transaction times and separate transfers on MAXTIME value in micro   */
/* seconds - 100 seconds seems to be enough                                 */

#define MAXTIME 100000000

/* The following definitions are made to make the task of parsing the       */
/* tcpdump lines easier.  A define must be made for the domain name of all  */
/* systems which might show up in the tcpdump log.                          */

/* server domain names */

#define HOST1 "newbury"

/* client domain names */

#define HOST2 "sunlight"

/* Define function prototypes */

void count_pkts_bytes();
void time_line();
void init();
void initfile();
void get_next_file(FILE *, int);
void critpath(long,long);
void results();
int compare1(const void *first, const void *second);
int compare2(const void *first, const void *second);
int compare3(const void *first, const void *second);
int compare4(const void *first, const void *second);
int compare5(const void *first, const void *second);
int compare6(const void *first, const void *second);
int compare7(const void *first, const void *second);
int compare8(const void *first, const void *second);
int compare9(const void *first, const void *second);
int compare10(const void *first, const void *second);

/* Define structure for "numeric short hand" for tcpdump line               */

typedef struct tdump {
  int pkttype;                /* packet type                                */
  int bytes;                  /* number of bytes in packet                  */
  int host;                   /* set to 1 if host is remote system (HOST2)  */
  long mss;                   /* advertised segment size                    */
  long win;                   /* receive window size in bytes               */
  long x1;                    /* x coordinate for host in plot              */
  long ipid;                  /* Packeet IP identification number           */
  long rport;                 /* Remote client port number                  */
  long ysec;                  /* seconds time stamp                         */
  long yusec;                 /* micro seconds time stamp                   */
  long ylong;                 /* long value of time stamp for sorting...    */
  unsigned long seqn;         /* sequence number of packet                  */
} tdump;

/* Define the structure for the xgraph output array                         */

typedef struct tline {
  int pkttype;                /* packet type                                */
  int origpkt;                /* stores original packet type (never a drop) */
  int host;                   /* set to 1 if host is remote system (HOST2)  */
  int bytes;                  /* number of bytes in this packet             */
  long mss;                   /* advertised segment size                    */
  long ipid;                  /* packet IP identification number            */
  long win;                   /* receive window size in bytes               */
  long x1;                    /* x coordinate for first host in plot        */
  float y1;                   /* relative time augmented for graphing       */
  long y1long;                /* value of y1 * 1e6 for sorting purposes     */
  long y2long;                /* value of y2 * 1e6 for sorting purposes     */
  long x2;                    /* x coordinate for second host in plot       */
  float y2;                   /* relative time augmented for graphing       */
  long rport;                 /* remote client port number                  */
  unsigned long seqn;         /* packet sequence number                     */
  int droptype;            /* Drop type:  4 = fast rexmit, 5 = time out     */
} tline;

/* Define the structure for the critical path analysis                      */

typedef struct cpath {
  int dlytype;              /* type of delay for a critical path segment    */
  float x1;                 /* x coordinate for plotting the critical path  */
  float y1;                 /* y coordinate for plotting the critical path  */
  float y2;                 /* second points offset from first point        */
  int src;                  /* source of delay: 1 if client, 0 if server    */
  long file;                /* the count of the current transfer            */
  int abort;                /* abort flag if transfer is in error somehow   */
  int status;               /* 1 = setup, 2 = GET/bulk-xfer, 3 = tear down  */
  int host;                 /* set to 1 if host is remote system (HOST2)    */
  long ipid;                  /* packet IP identification number            */
  long rport;                 /* remote client port number                  */
} cpath;

/* Define the structure for tracking critical path through the bulk         */
/* transfer phase of a HTTP transaction.                                    */

typedef struct xfrround {
  int phantom;                /* flag for phantom drop not seen in trace    */
  long data;               /* index of data packet associated with this ACK */
  long ack;                /* index of ack packet associated with a round   */
  long bytestrt;           /* start byte count of data dependant on this ACK*/
  long byteend;            /* end byte count of data dependant on this ACK  */
  int numpkts;             /* the number of data packets for round per TCP  */  
  int founddrop;           /* flag to indicate that a drop is in this round */
  int droptype;            /* Drop type:  4 = fast rexmit, 5 = time out     */
  long drop;               /* index of droped data packet on crit path      */
} xfrround;

/* Define globals */

long num_trans;             /* count of number of files which have cp analys*/
long num_omit;              /* number of transactions omitted from cp analys*/
long totrst;                /* total number of reset packets in trace       */
long totsyn;                /* total number of syn packets in trace         */
long totfin;                /* total number of fin packets in trace         */
long totack;                /* total number of ack packets in trace         */
long totdata;               /* total number of data packets in trace        */
long totdrops;              /* total number of dropped packets in trace     */
long num_intv;              /* total number of intervals in trace           */
long num_out;               /* total number of packets in time line output  */
long file_cnt;              /* total number of files transmitted            */
long num_tln;               /* total number of entries in time line array   */
long cur_cp;                /* pointer to current line in cp array          */
long pktcount[MAXTRACE];    /* count of number of files transfered in intrvl*/
long bytcount[MAXTRACE];    /* count of number of bytes transfered in intrvl*/
struct tdump dp[MAXTRACE];  /* tcpdump short hand array                     */
struct tline outl[MAXTRACE]; /* time line output in xgraph format           */
struct cpath cp[MAXTRACE];  /* critical path line array                     */
struct xfrround xfr[MAXSEQ]; /* bulk transfer rounds critical path          */ 
/****************************************************************************/
/* Main simply reads the command line and starts off the various functions  */
/****************************************************************************/

int main(int argc, char *argv[])
{

  FILE *fp1, *fp2;

  /* First initialize variables and open the input files                    */
  
  init();

  if(argc < 3)
    {
      printf("Usage: pbtcpeval {tcpdump file1} {tcpdump file2}\n");
      exit(0);
    }

  if((fp1 = fopen(argv[1], "r")) == NULL)
    {
      printf("Unable to open file %s\n", argv[1]);
      exit(0);
    }
  if((fp2 = fopen(argv[2], "r")) == NULL)
    {
      printf("Unable to open file %s\n", argv[2]);
      exit(0);
    }

  /* Now get the data and generate the results - giddy yap!                 */

  get_next_file(fp1, 0);
  count_pkts_bytes();
  get_next_file(fp2, 1);
  time_line();

  /* Now close the input files and print out the results files.             */

  fclose(fp1);
  fclose(fp2);

  results();

return 0;

/* End Main */
}

/****************************************************************************/
/* Generating the time line is mainly function of sorting the tline array   */
/* in order to group by port number (which determines breaks between file   */
/* transfers) packet type (so we can display different color lines in       */
/* xgraph for the different packet types), IP ID (so we know which lines    */
/* to connect and when there is a packet drop) and transfer times (so we    */
/* the order in which to print the.  Qsort is used to facilitate the        */
/* groupings.  NOTE:  The grouping by port number could be a problem if     */
/* port numbers are reused.  NOTE:  At this time, when a RESET is sent for  */
/* a transfer, this transfer is not considered in the results data.  Only   */
/* the number of RESETS are counted and output at this time.                */
/****************************************************************************/

void time_line()
{
  int setfirst, RST, lasthost;
  long i, j, k, l, lastport, minsec, minusec, maxsec, maxusec, stopflag;
  long lasttime, lastip;
  long lasttln;
  float holder;

  num_trans = 1;

  for(i=0;i<num_out;i++)
    if(dp[i].rport == 0) printf("START found rport = 0 at %ld\n",i);

  /* Now just generate a non-decimal version of the time stamp for the  */
  /* the purpose of sorting the DP array.  We must find the min time    */
  /* and subtract all times from it since otherwise the time is too     */
  /* large to hold in a long.                                           */

  minsec = dp[0].ysec;
  minusec = dp[0].yusec;
  for(i=0;i<num_out;i++) {
    if((dp[i].ysec < minsec) || (dp[i].ysec == minsec && dp[i].yusec < minusec))
      {
	minsec = dp[i].ysec;
	minusec = dp[i].yusec;
      }
  }
  for(i=0;i<num_out;i++) {
    dp[i].ylong = (dp[i].ysec-minsec+(dp[i].yusec-minusec)/1e6) * 1e6;
  }

  /* Now sort dp array keyed by time stamp and rport.  We sort by time and   */
  /* rport since we really want to group by rport but ports can be reused.   */

  qsort(dp,num_out,sizeof(dp[0]),compare10); /* sorting dp array by port #   */

  /*for(i=0;i<num_out;i++) {
        printf("Time = %ld Port = %ld Type = %d Host = %d ID = %ld \n",
	   dp[i].ylong,dp[i].rport,dp[i].pkttype,dp[i].host,dp[i].ipid);
	   }*/

  /* Now chunk through tdump array and find the break points between files  */
  /* tracking the min and max timestamps within a file.  After the indexes  */
  /* for a specific transfer have been found, sort the tdump array for the  */
  /* file by ip id numbers so that packet pairs can be matched and drops    */
  /* can be found.  Then, load the tline array with packet pairs for each   */
  /* file.                                                                  */

  j = 0;
  stopflag = 0;
  num_tln = 0;
  lastport = dp[0].rport;
  lasttime = dp[0].ylong;
  minsec = maxsec = dp[0].ysec;
  minusec = maxusec = dp[0].yusec;
  for(i=0;i<num_out;i++)
    {  /* find the end of a file transfer by port or by time */
      if(dp[i].rport != lastport || i == num_out - 1 || 
	 dp[i].ylong - lasttime > MAXTIME)
	{  /* now sort by ip id in order to match pairs for a file transfer */
	  if(i == num_out - 1) i++; /* Include the last packet in the file! */
	  qsort(&dp[stopflag],i-stopflag,sizeof(dp[0]),compare2); 
	  lastip = -1;
	  lasthost = dp[stopflag].host;
	  setfirst = 0;
	  lasttln = num_tln;
	  RST = 0;
	  for(k=stopflag;k<i;k++)  /* Cycle through data for a single file  */
	    {
	      /* enter data in tline for new pkt   */
	      if(lastip != dp[k].ipid || lasthost != dp[k].host)
		{
		  /* Since both hosts can use the same ipid's we must sort  */
		  /* each group of ipid's that are the same by host         */
		  l = k + 1;
		  while(dp[k].ipid == dp[l].ipid && l < i) l++;
		  if(l > k + 2)
		    qsort(&dp[k],l-k,sizeof(dp[0]),compare8);
		  lastip = dp[k].ipid;
		  lasthost = dp[k].host;
		  if(setfirst == 1) /* was pkt dropped?*/
		    {  
		      outl[num_tln].pkttype = 6;  /* set pkttype to drop    */
		      outl[num_tln].y2 = outl[num_tln].y1;
		      outl[num_tln].y2long = outl[num_tln].y2 * 1e6;
		      outl[num_tln].x2 = j + 1;
		      totdrops++;
		      num_tln++;
		    }
		  outl[num_tln].rport = dp[k].rport;
		  outl[num_tln].host = dp[k].host;
		  outl[num_tln].pkttype = dp[k].pkttype;
		  outl[num_tln].origpkt = dp[k].pkttype;
		  outl[num_tln].win = dp[k].win;
		  outl[num_tln].mss = dp[k].mss;
		  outl[num_tln].ipid = dp[k].ipid;
		  outl[num_tln].bytes = dp[k].bytes;
		  outl[num_tln].y1 = dp[k].ysec-minsec+(dp[k].yusec-minusec)/1e6;
		  outl[num_tln].y1long = outl[num_tln].y1 * 1e6;
		  outl[num_tln].seqn = dp[k].seqn;
		  outl[num_tln].x1 = j;
		  outl[num_tln].x2 = j + 1;
		  if(dp[k].pkttype == 8) {
		    RST = 1;
		    totrst++;
		  }
		  setfirst = 1;
		} else {
		  outl[num_tln].y2 = dp[k].ysec-minsec+(dp[k].yusec-minusec)/1e6;
		  outl[num_tln].y2long = outl[num_tln].y2 * 1e6;
		  /* Be sure that if host = remote that pkt start time is   */
                  /* in y1, and if host = server then pkt start time is y2  */
		  if(!outl[num_tln].host && outl[num_tln].y1 < outl[num_tln].y2) {
		    holder = outl[num_tln].y1;
		    outl[num_tln].y1 = outl[num_tln].y2;
		    outl[num_tln].y2 = holder;
		    outl[num_tln].y2long = outl[num_tln].y2 * 1e6;
		    outl[num_tln].y1long = outl[num_tln].y1 * 1e6;
		  }
		  if(outl[num_tln].host && outl[num_tln].y2 < outl[num_tln].y1) {
		    holder = outl[num_tln].y1;
		    outl[num_tln].y1 = outl[num_tln].y2;
		    outl[num_tln].y2 = holder;
		    outl[num_tln].y2long = outl[num_tln].y2 * 1e6;
		    outl[num_tln].y1long = outl[num_tln].y1 * 1e6;
		  }
		  setfirst = 0;
		  num_tln++;
		}
	    }

	  /* If there was a RESET flag set in any of the packets of this    */
	  /* transfer, then do not consider this transfer in the CP results */

	  if(!RST) {  /* calculate the delay critical path for the file      */
	    critpath(lasttln, num_tln);
	    printf("Completed file # %ld\n",num_trans);
	    num_trans++;
	  }

	  /* now add two final entries to the tline array which draw the    */
	  /* vertical lines between x-axis values to max time which frame   */
	  /* each file transfer.  These framing lines will have type = 7    */

	  outl[num_tln].rport = lastport;
	  outl[num_tln].x1 = j;
	  outl[num_tln].y1 = 0.0;
	  outl[num_tln].x2 = j;
	  outl[num_tln].y2 = maxsec - minsec + (maxusec - minusec)/1e6;
	  outl[num_tln].pkttype = 7;
	  num_tln++;
	  outl[num_tln].rport = lastport;
	  outl[num_tln].x1 = j + 1;
	  outl[num_tln].y1 = 0.0;
	  outl[num_tln].x2 = j + 1;
	  outl[num_tln].y2 = maxsec - minsec + (maxusec - minusec)/1e6;
	  outl[num_tln].pkttype = 7;
	  num_tln++;

	  stopflag = i;
	  j += 2;
	  file_cnt++;
	  lastport = dp[i].rport;
	  lasttime = dp[i].ylong;
	  minsec = maxsec = dp[i].ysec;
	  minusec = maxusec = dp[i].yusec;
	} else {   /* so, find min and max timestamps for this file         */
	  if((dp[i].ysec < minsec) || (dp[i].ysec == minsec && dp[i].yusec < minusec))
	    {
	      minsec = dp[i].ysec;
	      minusec = dp[i].yusec;
	    }
	  if((dp[i].ysec > maxsec) || (dp[i].ysec == maxsec && dp[i].yusec > maxusec))
	    {
	      maxsec = dp[i].ysec;
	      maxusec = dp[i].yusec;
	    }
	}
    }

/* End function time_line */
}

/****************************************************************************/
/* Parse out the information from the transfer of one file from each of the */
/* input files - this is a numeric short hand for the tcpdump files.        */
/* Assume that each line in the tcpdump files has the following format:     */
/*                                                                          */
/* hr:min:sec.usec {src adr}.{prt #} > {dest adr}.{prt #}: {pkttyp} {stuff} */
/*                                                                          */
/* and {stuff} has one of two formats:                                      */
/*                                                                          */
/*          ack {seq #} win {winsz} {tcpopts} (ttl {time}, id {ip id})      */
/*    or                                                                    */
/*          {seq #}:{seq #}(seg size) ack ...                               */
/*                                                                          */
/* All of the parsing is set up to handle these formats.  If the format is  */
/* different, then the program prints out an error message.                 */
/*                                                                          */
/****************************************************************************/

void get_next_file(FILE *fptr, int infile)
{
  int i, foundport, foundtype, killpacket;
  long hr, mn, sc, us;
  char j1[4], src[40], dstn[40], pkttyp[4], hardpart[200], bytebuff[10];
  char winsiz[] = "win";
  char msssiz[] = "mss";
  char syntyp[] = "S";
  char fintyp[] = "F";
  char fintyp1[] = "FP";
  char dottyp[] = ".";
  char pushtyp[] = "P";
  char resettyp[] = "R";
  char leftprn[] = "(";
  char rghtprn[] = ")";
  char ipidptr[] = "id";
  char colon[] = ":";
  char space[] = " ";
  char gtsign[] = ">";
  char *tokenPrt;
  char *lineptr1;
  char *lineptr2;
  int hrcnt = 0;
  int oldflag = 0;

  bzero(src,40);
  bzero(dstn,40);
  bzero(pkttyp,1);
  bzero(hardpart,200);

  /* First parse out the time stamp, src and dest names, packet type. Then  */
  /* get the rest of the string which will be further parsed below.         */

  while(fscanf(fptr, "%ld:%ld:%ld.%ld %s %s %s %s", 
	       &hr, &mn, &sc, &us, src, j1, dstn, pkttyp) != EOF)
    {  

      /* set the x1 value in the xgrph array to indicate whether data came  */
      /* from the first or the second input file.                           */

      dp[num_out].x1 = infile;

      /* now load the rest of the tcpdump line into the string hardpart     */

      i = 0;
      while(!feof(fptr) && (hardpart[i] = fgetc(fptr)) != '\n') i++;

      if(feof(fptr)) break;

      /* first add the time stamp to the xgrph array as a float value       */
      /* Be sure that time is always increasing ie. go from 23:59:59 to     */
      /* 24:00:00 and not 00:00:00...                                       */

      if(hr == 23 && (oldflag == 0 || oldflag == 2)) oldflag = 1;
      if(hr == 0 && oldflag == 1) {
	oldflag = 2;
	hrcnt += 24;
      }
      dp[num_out].ysec = 3600 * (hr + hrcnt) + 60 * mn + sc;
      dp[num_out].yusec = us;

      if(DEBUG==1) printf("Time stamp = %ld sec, %ld usec\n", 
		       dp[num_out].ysec,dp[num_out].yusec);

      /* Now we parse out the host name and port numbers of the source      */
      /* and destination systems.  In particular, we must find the port #   */
      /* of the remote system since that is how we can most easily determine*/
      /* the end of an individual file transfer (for HTTP/1.0 anyway).      */
      
      foundport = 0;
      killpacket = 0;
      tokenPrt = strtok(src, ".");
      if(strstr(tokenPrt, HOST2) != NULL)
	{
	  
	  for(i=0;i<4;i++) tokenPrt = strtok(NULL, ".");
	  dp[num_out].rport = strtol(tokenPrt, NULL, 10);
	  dp[num_out].host = 1;
	  foundport = 1;
	}
      tokenPrt = strtok(dstn, ".");
      if(strstr(tokenPrt, HOST2) != NULL)
	{
	  for(i=0;i<4;i++) tokenPrt = strtok(NULL, ".");
	  dp[num_out].rport = strtol(tokenPrt, NULL, 10);
	  foundport = 1;
	}
      if(!foundport || dp[num_out].rport == 0) 
	{
	  printf("Unable to find port number - discarding packet\n");
	  killpacket = 1;
	} else {
	  if(DEBUG==1) printf("Remote host port = %ld\n", dp[num_out].rport);
	}
      /* Now determine the packet type.  Save as SYN = 1, ACK = 2, DATA = 3 */
      /* FIN = 4, OTHER = 5, RESET = 8.  This simple set up does not        */
      /* consider piggy backed ACKS at the present time.                    */

      foundtype = 0;
      if(strstr(pkttyp, syntyp) != NULL)
	{
	  dp[num_out].pkttype = 1;
	  foundtype = 1;
	}
      if(strstr(pkttyp, fintyp) != NULL || strstr(pkttyp, fintyp1) != NULL)
	{
	  dp[num_out].pkttype = 4;
	  foundtype = 1;
	}
      if(strstr(pkttyp, resettyp) != NULL)
	{
	  dp[num_out].pkttype = 8;
	  foundtype = 1;
	}
      if(strstr(pkttyp, dottyp) != NULL)
	{
	  if(isalpha(hardpart[1]))
	    {
	      dp[num_out].pkttype = 2;
	    } else {
	      dp[num_out].pkttype = 3;
	    }
	  foundtype = 1;
	}
      if(strstr(pkttyp, pushtyp) != NULL && dp[num_out].pkttype != 4)
	{
	  dp[num_out].pkttype = 3;
	  foundtype = 1;
	}
      if(!foundtype) dp[num_out].pkttype = 5;

      if(DEBUG==1) printf("Packet type = %d\n",dp[num_out].pkttype); 

      /* Now grab the sequence number from packets.  If the packet iss an    */
      /* ack, the seq # is the next acknowledged byte number.  If it is any  */
      /* other type packet, the seq # is the last sent byte number.          */

      lineptr1 = hardpart;
      bzero(bytebuff,10);
      if(dp[num_out].pkttype != 2)    /* Consider non-ack packets first   */
	{                         /* look for first ":" and match with " "   */
	 if((lineptr1 = strstr(lineptr1, colon)) != NULL) 
	   {
	     lineptr2 = lineptr1;
	     if((lineptr2 = strstr(lineptr2, leftprn)) != NULL)
	       {
		 strncpy(bytebuff,lineptr1+1,lineptr2 - lineptr1 - 1);
		 dp[num_out].seqn = strtoul(bytebuff, NULL, 10);
		 if(DEBUG==1) printf("Sequence number  = %lu\n",dp[num_out].seqn);
	       }

	   }
	} else {                   /* Now consider ack packets               */
	  lineptr1 += 5;           /* put pointer at first digit of seq #    */
	  lineptr2 = lineptr1;
	  if((lineptr2 = strstr(lineptr2, space)) != NULL)
	    {
	      strncpy(bytebuff,lineptr1,lineptr2 - lineptr1);
	      dp[num_out].seqn = strtoul(bytebuff, NULL, 10);
	      if(DEBUG==1) printf("Sequence number  = %lu\n", dp[num_out].seqn);
	    }
	}

      /* Now grab the number of bytes in the packet.  This appears in parens */
      /* after the sequence number if the packet is a data packet.           */

      if(dp[num_out].pkttype == 3)    /* Only consider data packets          */
	{
	  lineptr1 = hardpart;     /* add bytes to byte count for current bin*/
	  lineptr2 = hardpart;
	  bzero(bytebuff,10);      /* look for first "(" and match with ")"  */
	  if((lineptr1 = strstr(lineptr1, leftprn)) != NULL) 
	    {
	      if((lineptr2 = strstr(lineptr2, rghtprn)) != NULL)
		{
		 strncpy(bytebuff,lineptr1+1,lineptr2 - lineptr1 - 1);
		 dp[num_out].bytes = strtol(bytebuff, NULL, 10);
		 if(DEBUG==1) printf("Byte count  = %d\n", dp[num_out].bytes);
		}
	    } else {
	      printf("Packet size could not be found for data packet\n");
	    }
	}

      /* Now grab the window size of the packet.  This appears following the */
      /* win string and should appear in every packet.                       */

      if(dp[num_out].pkttype == 3)    /* Only consider data packets          */
	{
	  lineptr1 = hardpart;     /* add bytes to byte count for current bin*/
	  bzero(bytebuff,10);      /* look for "win" match in hardpart       */
	  if((lineptr1 = strstr(lineptr1, winsiz)) != NULL) 
	    {
	      lineptr2 = lineptr1 + 4;
	      if((lineptr2 = strstr(lineptr2, space)) != NULL)
		{
		  strncpy(bytebuff,lineptr1+3,lineptr2 - lineptr1 - 1);
		  dp[num_out].win = strtol(bytebuff, NULL, 10);
		  if(DEBUG==1) printf("Win size  = %ld\n", dp[num_out].win);
		}
	    } else {
	      printf("Window size could not be found for data packet\n");
	    }
	}

      /* Now grab the min seg size of the packet. This appears following the */
      /* mss string and should appear in the SYN packets                     */

      if(dp[num_out].pkttype == 1)    /* Only consider syn packets           */
	{
	  lineptr1 = hardpart;     /* add bytes to byte count for current bin*/
	  bzero(bytebuff,10);      /* look for "mss" match in hardpart       */
	  if((lineptr1 = strstr(lineptr1, msssiz)) != NULL) 
	    {
	      lineptr2 = lineptr1 + 4;
	      if((lineptr2 = strstr(lineptr2, gtsign)) != NULL)
		{
		  strncpy(bytebuff,lineptr1+3,lineptr2 - lineptr1 - 1);
		  dp[num_out].mss = strtol(bytebuff, NULL, 10);
		  if(DEBUG==1) printf("Mss size  = %ld\n", dp[num_out].mss);
		}
	    } else {
	      printf("Min seg size could not be found for data packet\n");
	    }
	}

      /* Now grab the ip identification for this packet.  This appears in    */
      /* parens at the end of the packet along with TTL info.                */
      
      lineptr1 = hardpart;
      bzero(bytebuff,10);  /* look for "id" and match with ")"               */
      if((lineptr1 = strstr(lineptr1, ipidptr)) != NULL)
	{
	  lineptr2 = lineptr1;
	  if((lineptr2 = strstr(lineptr2, rghtprn)) != NULL)
	    {
	      strncpy(bytebuff,lineptr1+3,lineptr2 - lineptr1 - 3);
	      dp[num_out].ipid = strtol(bytebuff, NULL, 10);
	    }
	} else {
	  printf("IP ID not found in packet - packet will be discarded\n");
	  killpacket = 1;
	}
      if(DEBUG==1) printf("IP ID  = %ld\n", dp[num_out].ipid);

      /* increment the output array counter and reset input buffers        */

      bzero(src,40);
      bzero(dstn,40);
      bzero(pkttyp,1);
      bzero(hardpart,200);

      if(killpacket) 
	{
	  killpacket = 0;
	} else {
	  num_out++;
	}
      if(num_out > MAXTRACE)
	{
	  printf("Buffers filled before end of scan. Increase MAXTRACE\n");
	  exit(1);
	}
    }

/* End function get_next_file */
}

/****************************************************************************/
/* This function counts up the number of packets and number of bytes that   */
/* are transferred in specified intervals of time (USINTRVL and SECINTRVL). */
/* Time starts with the first transfer and data transferred is only         */
/* calculated for ONE of the input files (the first one specified on the    */
/* command line.  This goes for the print out of the counts of the packet   */
/* types as well.                                                           */
/****************************************************************************/

void count_pkts_bytes()
{

  long i, basse, basus, curse, curus, old_intv, lastipid;
  long diff, usintrvl, secintrvl;
  int start, flag;

  start = 0;
  lastipid = 0;
  
  for(i=0;i<num_out;i++)
    {
      /* Now count packets which appear in each time interval bin and be    */
      /* be sure not to double count packets.                               */

      if(start == 0) 
	{
	  basse = dp[i].ysec;      /* Set base time for the bins start   */
	  basus = dp[i].yusec;
	  if(DEBUG==2) printf("Start Basse %ld Basus %ld\n",basse,basus);
	  pktcount[num_intv]++;
	  start = 1;
	} else {
	  if(lastipid != dp[i].ipid)
	    {
	      curse = basse = dp[i].ysec;
	      curus = basus = dp[i].yusec;
	      diff = 0;
	      usintrvl = curus - basus;
	      if(usintrvl < 0) 
		{
		  usintrvl = 1000000 + curus - basus;
		  diff = 1;
		}
	      secintrvl = curse - basse  - diff;
	      if(secintrvl > SECINTRVL)
		{
		  flag = 1;
		} else {
		  if((secintrvl == SECINTRVL) && (usintrvl > USINTRVL))
		    { 
		      flag = 1;
		    } else {
		      if((SECINTRVL == 0) && (usintrvl > USINTRVL))
			{
			  flag = 1;
			} else {
			  flag = 0;
			}
		    }
		}
	      if(flag == 0)                 /* packet is in current bin */
		{
		  pktcount[num_intv]++;
		  if(DEBUG==2) printf("Packet count of bin %ld = %ld\n", 
				   num_intv, pktcount[num_intv]);
		} else {                 /* packet isn't in current bin */
		  old_intv = num_intv;   /* find correct bin for packet */
		  if(SECINTRVL == 0)     /* assumes USINTRVL > 0        */
		    {
		      num_intv += secintrvl * 1000000/USINTRVL;
		      if(usintrvl > USINTRVL)  
			num_intv += (usintrvl - USINTRVL)/USINTRVL + 1;
		    } else {
		      num_intv += (secintrvl - SECINTRVL)/SECINTRVL + 1;
		    }
		  if(old_intv == num_intv) num_intv++;
		  for(i=0;i<(num_intv - old_intv);i++)/* update bin markers */
		    {
		      basus += USINTRVL;
		      if(basus > 1000000) 
			{
			  basse++;
			  basus -= 1000000;
			}
		      basse += SECINTRVL;
		    }
		  pktcount[num_intv]++;
		  if(DEBUG==2) printf("New Basse = %ld Basus = %ld\n", 
				   basse, basus);
		  if(DEBUG==2) printf("Total Byte count last interval = %ld\n",
				   bytcount[num_intv-1]);
		}
	    }
	}
      
      /* Determine packet type and add byte counts to each interval bin -   */
      /* don't double count!                                                */

      if(lastipid != dp[i].ipid) {
	bytcount[num_intv] += dp[i].bytes;
	switch (dp[i].pkttype) {
	case 1:
	  totsyn++;
	  break;
	case 2:
	  totack++;
	  break;
	case 3:
	  totdata++;
	  break;
	case 4:
	  totfin++;
	  break;
	}
      lastipid = dp[i].ipid;
      }
    }

  printf("Total syn packets = %ld\n", totsyn);
  printf("Total fin packets = %ld\n", totfin);
  printf("Total ack packets = %ld\n", totack);
  printf("Total data packets = %ld\n", totdata);

/* End function count_pkt_bytes */
}

/****************************************************************************/
/* Init initializes variables and array used in this program                */
/****************************************************************************/

void init()
{
  long i;

  num_omit = 0;
  totsyn = 0;
  totfin = 0;
  totack = 0;
  totdata = 0;
  totdrops = 0;
  totrst = 0;
  num_intv = 0;
  num_out = 0;
  num_tln = 0;
  file_cnt = 0;
  cur_cp = 0;
  for(i=0;i<MAXTRACE;i++)
    {
      pktcount[i] = 0;
      bytcount[i] = 0;
      dp[i].seqn = 0;
      dp[i].bytes = 0;
      dp[i].x1 = 0;
      dp[i].ipid = 0;
      dp[i].rport = 0;
      dp[i].pkttype = 0;
      dp[i].ysec = 0;
      dp[i].yusec = 0;
      dp[i].ylong = 0;
      dp[i].host = 0;
      dp[i].win = 0;
      dp[i].mss = 0;
      outl[i].bytes = 0;
      outl[i].mss = 0;
      outl[i].win = 0;
      outl[i].rport = 0;
      outl[i].x1 = 0;
      outl[i].x2 = 0;
      outl[i].y1 = 0.0;
      outl[i].y1long = 0;
      outl[i].y2 = 0.0;
      outl[i].pkttype = 0;
      outl[i].origpkt = 0;
      outl[i].seqn = 0;
      outl[i].host = 0;
      outl[i].droptype = 0;
      cp[i].dlytype = 0;
      cp[i].x1 = 0.0;
      cp[i].y1 = 0.0;
      cp[i].y2 = 0.0;
      cp[i].src = 0;
      cp[i].file = 0;
      cp[i].abort = 0;
      cp[i].status = 0;
      cp[i].host = 0;
      cp[i].ipid = 0;
      cp[i].rport = 0;
    }

/* End function init */
}

/****************************************************************************/
/* Write out the files which contain the time line and the count data       */
/* There are lots and lots of results!!!                                    */
/****************************************************************************/

void results()
{
  int startflag, startflag1, curcnt, curcnt1, seccnt, nodrp, cntwodrps, numback;
  float mean, varval, mint, maxt, mean1, varval1, mint1, maxt1, mwodrp, xmarker;
  float cpcli, cpsrv, cpnet, cptmo, cpdrp, mean2, varval2, mean3, varval3;
  float mean4, varval4, mean5, varval5, mean6, varval6, mean7, varval7;
  float mean8, varval8, mean9, varval9, mean10, varval10, mean11, varval11;
  float mean12, varval12, mean13, varval13, mean14, varval14, strttime, mint2;
  long i, j, valcnt, dropcnt, filedropcnt, valcnt1, dropcnt1, lastport;
  long cpcnt, cpcnt1, cpcnt2, cpcnt3, cpcnt4, cpcnt5, cpcnt6, cpcnt7, backcnt;
  long cpcnt8, cpcnt9;
  FILE *fp1, *fp3, *fp4, *fp5, *fp6, *fp7, *fp8, *fp9, *fp10, *fp11, *fp12, *fp13;
  FILE *fp14, *fp15, *fp16, *fp17;

  seccnt = 0;
  printf("Total reset packets = %ld\n", totrst);
  printf("Total drop packets = %ld\n", totdrops);

  /* The time line data must be generated carefully.  We want to generate   */
  /* a file which splits out the points by packet type.  This requires us   */
  /* to sort the tline structure using packet type as the key.              */

  qsort(outl,num_tln,sizeof(outl[0]),compare3);

  /*  for(i=0;i<num_tln;i++) printf("%ld %f %ld %f %d\n", outl[i].x1, outl[i].y1,
			      outl[i].x2, outl[i].y2, outl[i].pkttype); */

  if(GRAPHTOOL == 0) /* GRAPHTOOL = 0 means generate files for gnuplot     */
    {
      if((fp3 = fopen("syn.tln", "w")) == NULL)
	{
	  perror("Unable to open syn.tln\n");
	  exit(1);
	}
      if((fp4 = fopen("ack.tln", "w")) == NULL)
	{
	  perror("Unable to open ack.tln\n");
	  exit(1);
	}
      if((fp5 = fopen("data.tln", "w")) == NULL)
	{
	  perror("Unable to open data.tln\n");
	  exit(1);
	}
      if((fp6 = fopen("fin.tln", "w")) == NULL)
	{
	  perror("Unable to open fin.tln\n");
	  exit(1);
	}
      if((fp7 = fopen("oth.tln", "w")) == NULL)
	{
	  perror("Unable to open oth.tln\n");
	  exit(1);
	}
      if((fp8 = fopen("drop.tln", "w")) == NULL)
	{
	  perror("Unable to open drop.tln\n");
	  exit(1);
	}
      if((fp9 = fopen("out.tln", "w")) == NULL)
	{
	  perror("Unable to open out.tln\n");
	  exit(1);
	}
      for(i=0;i<num_tln;i++)
	{
	  switch(outl[i].pkttype) {
	  case 1:
	    fprintf(fp3,"%ld %f\n", outl[i].x1, outl[i].y1);
	    fprintf(fp3,"%ld %f\n", outl[i].x2, outl[i].y2);
	    fprintf(fp3,"\n");
	    break;
	  case 2:
	    fprintf(fp4,"%ld %f\n", outl[i].x1, outl[i].y1);
	    fprintf(fp4,"%ld %f\n", outl[i].x2, outl[i].y2);
	    fprintf(fp4,"\n");
	    break;
	  case 3:
	    fprintf(fp5,"%ld %f\n", outl[i].x1, outl[i].y1);
	    fprintf(fp5,"%ld %f\n", outl[i].x2, outl[i].y2);
	    fprintf(fp5,"\n");
	    break;
	  case 4:
	    fprintf(fp6,"%ld %f\n", outl[i].x1, outl[i].y1);
	    fprintf(fp6,"%ld %f\n", outl[i].x2, outl[i].y2);
	    fprintf(fp6,"\n");
	    break;
	  case 5:
	    fprintf(fp7,"%ld %f\n", outl[i].x1, outl[i].y1);
	    fprintf(fp7,"%ld %f\n", outl[i].x2, outl[i].y2);
	    fprintf(fp7,"\n");
	    break;
	  case 6:
	    fprintf(fp8,"%ld %f\n", outl[i].x1, outl[i].y1);
	    fprintf(fp8,"%ld %f\n", outl[i].x2, outl[i].y2);
	    fprintf(fp8,"\n");
	    break;
	  case 7:
	    fprintf(fp9,"%ld %f\n", outl[i].x1, outl[i].y1);
	    fprintf(fp9,"%ld %f\n", outl[i].x2, outl[i].y2);
	    fprintf(fp9,"\n");
	    break;
	  }
	}
      fclose(fp3);
      fclose(fp4);
      fclose(fp5);
      fclose(fp6);
      fclose(fp7);
      fclose(fp8);
      fclose(fp9);

    } else {  /* GRAPHTOOL = 1 means generate files for xgraph              */

      if((fp3 = fopen("xgraph.out", "w")) == NULL)
	{
	  perror("Unable to open xgraph.out\n");
	  exit(1);
	}
      for(i=0;i<num_tln;i++)
	{
	  switch(outl[i].pkttype) {
	  case 1:
	    if(seccnt == 0)
	      {
		fprintf(fp3,"\"SYN\"\n");
		seccnt++;
	      }
	    fprintf(fp3,"move %ld %f\n", outl[i].x1, outl[i].y1);
	    fprintf(fp3,"draw %ld %f\n", outl[i].x2, outl[i].y2);
	    break;
	  case 2:
	    if(seccnt == 1)
	      {
		fprintf(fp3,"\n\"ACK\"\n");
		seccnt++;
	      }
	    fprintf(fp3,"move %ld %f\n", outl[i].x1, outl[i].y1);
	    fprintf(fp3,"draw %ld %f\n", outl[i].x2, outl[i].y2);
	    break;
	  case 3:
	    if(seccnt == 2)
	      {
		fprintf(fp3,"\n\"DATA\"\n");
		seccnt++;
	      }
	    fprintf(fp3,"move %ld %f\n", outl[i].x1, outl[i].y1);
	    fprintf(fp3,"draw %ld %f\n", outl[i].x2, outl[i].y2);
	    break;
	  case 4:
	    if(seccnt == 3)
	      {
		fprintf(fp3,"\n\"FIN\"\n");
		seccnt++;
	      }
	    fprintf(fp3,"move %ld %f\n", outl[i].x1, outl[i].y1);
	    fprintf(fp3,"draw %ld %f\n", outl[i].x2, outl[i].y2);
	    break;
	  case 5:
	    if(seccnt <= 4)
	      {
		fprintf(fp3,"\n\"OTHER\"\n");
		seccnt = 5;
	      }
	    fprintf(fp3,"move %ld %f\n", outl[i].x1, outl[i].y1);
	    fprintf(fp3,"draw %ld %f\n", outl[i].x2, outl[i].y2);
	    break;
	  case 6:
	    if(seccnt <= 5)
	      {
		fprintf(fp3,"\n\"DROP\"\n");
		seccnt = 6;
	      }
	    fprintf(fp3,"move %ld %f\n", outl[i].x1, outl[i].y1);
	    fprintf(fp3,"draw %ld %f\n", outl[i].x2, outl[i].y2);
	    break;
	  case 7:
	    if(seccnt <= 6)
	      {
		fprintf(fp3,"\n");
		seccnt = 7;
	      }
	    fprintf(fp3,"move %ld %f\n", outl[i].x1, outl[i].y1);
	    fprintf(fp3,"draw %ld %f\n", outl[i].x2, outl[i].y2);
	    break;
	  }
	}
      fclose(fp3);
    }

  /* now write out file_delays.txt which is simply the total latency for    */
  /* each file transfer.  This should be close to the results for latency   */
  /* on the client but not exactly the same due to possible lost FIN or ACK */
  /* packets during connection tear down                                    */

  if((fp1 = fopen("delay_info.txt", "w")) == NULL)
    {
      perror("Unable to open delay_info.txt\n");
      exit(1);
    }

  if((fp3 = fopen("file_delay.txt", "w")) == NULL)
    {
      perror("Unable to open file_delays.txt\n");
      exit(1);
    }
  varval = 0.0;
  mean = 0.0;
  maxt = 0.0;
  mint = 0.0;
  valcnt = 0;
  startflag = 0;
  for(i=0;i<num_tln;i++)
    if(outl[i].pkttype == 7 && outl[i].x1%2)
      {
	fprintf(fp3,"%f\n",outl[i].y2);
	mean += outl[i].y2;
	valcnt++;
	if(startflag == 0) 
	  {
	    maxt = outl[i].y2;
	    mint = outl[i].y2;
	    startflag = 1;
	  } else {
	    if(mint > outl[i].y2) mint = outl[i].y2;
	    if(maxt < outl[i].y2) maxt = outl[i].y2;
	  }
      }
  fclose(fp3);
  mean = mean/valcnt;
  for(i=0;i<num_tln;i++)
    if(outl[i].pkttype == 7 && outl[i].x1%2)
	varval += (outl[i].y2 - mean) * (outl[i].y2 - mean);

  printf("Writing all transfer and delay info to delay_info.txt\n");
  fprintf(fp1,"Number of files transferred:  %ld\n", valcnt);
  fprintf(fp1,"Mean file latency:            %f\n", mean);
  fprintf(fp1,"Variance of file latency:     %f\n", varval/valcnt);
  fprintf(fp1,"Standard Dev of file latency: %f\n", sqrt(varval/valcnt));
  fprintf(fp1,"Maximum file latency:         %f\n", maxt);
  fprintf(fp1,"Minimum file latency:         %f\n", mint);
  fprintf(fp1,"\n");

  /* now write out the file pkt_delays.txt which is are the packet transit  */
  /* times as measured at the end points.  These should be close to the     */
  /* times measured using poip but not exactly due to packet sizes.         */
  /* Calculate delays in both directions.  If outl[x].host = 1 then packets */
  /* are FROM REMOTE!  Also write out pkt_delays1.txt which are the delays  */
  /* TO REMOTE.  Also print out summary stats for delays in both directions */

  if((fp3 = fopen("pkt_delays.txt", "w")) == NULL)
    {
      perror("Unable to open pkt_delays.txt\n");
      exit(1);
    }
  if((fp4 = fopen("pkt_delays1.txt", "w")) == NULL)
    {
      perror("Unable to open pkt_delays1.txt\n");
      exit(1);
    }
  varval = varval1 = 0.0;
  mean = mean1 = 0.0;
  maxt = maxt1 = 0.0;
  mint = mint1 = 0.0;
  valcnt = valcnt1 = 0;
  startflag = startflag1 = 0;
  for(i=0;i<num_tln;i++)
    if(outl[i].pkttype != 6 && outl[i].pkttype != 7) {
      if(outl[i].host) /* if host = 1 then transfer is FROM REMOTE */
	{
	  if(outl[i].y1 > outl[i].y2)
	    {
	      fprintf(fp3,"%f\n",outl[i].y1 - outl[i].y2);
	      mean += outl[i].y1 - outl[i].y2;
	      valcnt++;
	      if(startflag == 0 && outl[i].y1 - outl[i].y2 > 0.0) 
		{
		  maxt = outl[i].y1 - outl[i].y2;
		  mint = outl[i].y1 - outl[i].y2;
		  startflag = 1;
		} else {
		  if(mint > outl[i].y1 - outl[i].y2 && outl[i].y1 - outl[i].y2 > 0.0) 
		    mint = outl[i].y1 - outl[i].y2;
		  if(maxt < outl[i].y1 - outl[i].y2) maxt = outl[i].y1 - outl[i].y2;
		}
	    } else {
	      fprintf(fp3,"%f\n",outl[i].y2 - outl[i].y1);
	      mean += outl[i].y2 - outl[i].y1;
	      valcnt++;
	      if(startflag == 0 && outl[i].y2 - outl[i].y1 > 0.0) 
		{
		  maxt = outl[i].y2 - outl[i].y1;
		  mint = outl[i].y2 - outl[i].y1;
		  startflag = 1;
		} else {
		  if(mint > outl[i].y2 - outl[i].y1 && outl[i].y2 - outl[i].y1 > 0.0 ) 
		    mint = outl[i].y2 - outl[i].y1;
		  if(maxt < outl[i].y2 - outl[i].y1) maxt = outl[i].y2 - outl[i].y1;
		}
	    }
	} else { /* if host = 0 then transfer is TO REMOTE   */
	  if(outl[i].y1 > outl[i].y2)
	    {
	      fprintf(fp4,"%f\n",outl[i].y1 - outl[i].y2);
	      mean1 += outl[i].y1 - outl[i].y2;
	      valcnt1++;
	      if(startflag1 == 0 && outl[i].y1 - outl[i].y2 > 0.0) 
		{
		  maxt1 = outl[i].y1 - outl[i].y2;
		  mint1 = outl[i].y1 - outl[i].y2;
		  startflag1 = 1;
		} else {
		  if(mint1 > outl[i].y1 - outl[i].y2 && outl[i].y1 - outl[i].y2 > 0.0) 
		    mint1 = outl[i].y1 - outl[i].y2;
		  if(maxt1 < outl[i].y1 - outl[i].y2) maxt1 = outl[i].y1 - outl[i].y2;
		}
	    } else {
	      fprintf(fp4,"%f\n",outl[i].y2 - outl[i].y1);
	      mean1 += outl[i].y2 - outl[i].y1;
	      valcnt1++;
	      if(startflag1 == 0 && outl[i].y2 - outl[i].y1 > 0.0) 
		{
		  maxt1 = outl[i].y2 - outl[i].y1;
		  mint1 = outl[i].y2 - outl[i].y1;
		  startflag1 = 1;
		} else {
		  if(mint1 > outl[i].y2 - outl[i].y1 && outl[i].y2 - outl[i].y1 > 0.0) 
		    mint1 = outl[i].y2 - outl[i].y1;
		  if(maxt1 < outl[i].y2 - outl[i].y1) maxt1 = outl[i].y2 - outl[i].y1;
		}
	    }
	}
    }
  fclose(fp3);
  fclose(fp4);
  mean = mean/valcnt;
  mean1 = mean1/valcnt1;
  for(i=0;i<num_tln;i++)
    if(outl[i].pkttype != 6 && outl[i].pkttype != 7) {
      if(outl[i].host)
	{
	  if(outl[i].y1 > outl[i].y2)
	    {
	      varval += (outl[i].y1-outl[i].y2-mean)*(outl[i].y1-outl[i].y2-mean);
	    } else {
	      varval += (outl[i].y2-outl[i].y1-mean)*(outl[i].y2-outl[i].y1-mean);
	  }
	} else {
	  if(outl[i].y1 > outl[i].y2)
	    {
	      varval1 += (outl[i].y1-outl[i].y2-mean1)*(outl[i].y1-outl[i].y2-mean1);
	    } else {
	      varval1 += (outl[i].y2-outl[i].y1-mean1)*(outl[i].y2-outl[i].y1-mean1);
	    }
	}
    }
  printf("Packet Delay stats FROM REMOTE in pkt_delays.txt\n");
  fprintf(fp1,"Packet Delay stats FROM REMOTE (details in pkt_delays.txt)\n");
  fprintf(fp1,"Number of packets transferred:  %ld\n", valcnt);
  fprintf(fp1,"Mean packet latency:            %f\n", mean);
  fprintf(fp1,"Variance of packet latency:     %f\n", varval/valcnt);
  fprintf(fp1,"Standard Dev of packet latency: %f\n", sqrt(varval/valcnt));
  fprintf(fp1,"Maximum packet latency:         %f\n", maxt);
  fprintf(fp1,"Minimum packet latency:         %f\n", mint);
  fprintf(fp1,"\n");
  printf("Packet Delay stats TO REMOTE in pkt_delays1.txt\n");
  fprintf(fp1,"Packet Delay stats TO REMOTE (details in pkt_delays1.txt)\n");
  fprintf(fp1,"Number of packets transferred:  %ld\n", valcnt1);
  fprintf(fp1,"Mean packet latency:            %f\n", mean1);
  fprintf(fp1,"Variance of packet latency:     %f\n", varval1/valcnt1);
  fprintf(fp1,"Standard Dev of packet latency: %f\n", sqrt(varval1/valcnt1));
  fprintf(fp1,"Maximum packet latency:         %f\n", maxt1);
  fprintf(fp1,"Minimum packet latency:         %f\n", mint1);
  fprintf(fp1,"\n");

  /* now calculate and print out some basic packet loss characteristics     */
  /* These should be similar to those measured by poip.  this routine has a */
  /* nasty double loop in it which should be removed!!!!!!!  Just add       */
  /* another member to tline which stores the file latency.  Write out loss */
  /* information for drops in both directions and print out summary loss    */
  /* information for drops in both directions.  Also print out the file     */
  /* latency for files with and without drops.                              */

  varval = varval1 = 0.0;
  mean = 0.0;
  maxt = maxt1 = 0.0;
  dropcnt = dropcnt1 = 0;
  filedropcnt = 0;
  curcnt = curcnt1 = 1;
  lastport = -1;
  for(i=0;i<num_tln;i++)
    if(outl[i].pkttype == 6)
      {
	if(lastport == -1)
	  {
	    filedropcnt = 1;
	    lastport = outl[i].rport;
	    for(j=0;j<num_tln;j++)
	      if(outl[j].rport == lastport && outl[j].pkttype == 7 && outl[j].x1%2)
		{
		  mean += outl[j].y2;
		  break;
		}
	    if(outl[i].host)
	      {
		dropcnt++;
		maxt = 1;
	      } else {
		dropcnt1++;
		maxt1 = 1;
	      }
	  } else {
	    if(lastport != outl[i].rport)
	      {
		filedropcnt++;
		lastport = outl[i].rport;
		for(j=0;j<num_tln;j++)
		  if(outl[j].rport == lastport && outl[j].pkttype == 7 && outl[j].x1%2)
		    {
		      mean += outl[j].y2;
		      break;
		    }
		if(outl[i].host)
		  {
		    dropcnt++;
		    if(curcnt > maxt) maxt = curcnt;
		    curcnt = 1;
		  } else {
		    dropcnt1++;
		    if(curcnt1 > maxt1) maxt1 = curcnt1;
		    curcnt1 = 1;
		  }
	      } else {
		if(outl[i].host)
		  {
		    dropcnt++;
		    curcnt++;
		  } else {
		    dropcnt1++;
		    curcnt1++;
		  }
	      }
	  }
      }

  if((fp3 = fopen("file_delays_drp.txt", "w")) == NULL)
    {
      perror("Unable to open file_delays_drp.txt\n");
      exit(1);
    }
  if((fp4 = fopen("file_delays_ndrp.txt", "w")) == NULL)
    {
      perror("Unable to open file_delays_ndrp.txt\n");
      exit(1);
    }
  cntwodrps = 0;
  mwodrp = 0.0;
  lastport = -1;
  mean = mean/filedropcnt;
  for(i=0;i<num_tln;i++)
    if(outl[i].pkttype == 6 && lastport != outl[i].rport)
      {
	lastport = outl[i].rport;
	for(j=0;j<num_tln;j++)
	  if(outl[j].rport == lastport && outl[j].pkttype == 7 && outl[j].x1%2)
	    {
	      fprintf(fp3,"%f\n", outl[j].y2);
	      varval += (outl[j].y2 - mean) * (outl[j].y2 - mean);
	      break;
	    }
      } else {
	if(outl[i].pkttype == 7 && outl[i].x1%2)
	  {
	    nodrp = 1;
	    for(j=0;j<num_tln;j++) 
	      if(outl[j].pkttype == 6 && outl[j].rport == outl[i].rport)
		{
		  nodrp = 0;
		  break;
		}
	    if(nodrp)
	      {
		cntwodrps++;
		mwodrp += outl[i].y2;
		fprintf(fp4,"%f\n", outl[i].y2);
	      }
	  }
      }
  mwodrp = mwodrp/cntwodrps;
  for(i=0;i<num_tln;i++)
    if(outl[i].pkttype == 7 && outl[i].x1%2)
      {
	nodrp = 1;
	for(j=0;j<num_tln;j++) 
	  if(outl[j].pkttype == 6 && outl[j].rport == outl[i].rport) 
	    {
	      nodrp = 0;
	      break;
	    }
	if(nodrp) varval1 += (outl[i].y2 - mwodrp) * (outl[i].y2 - mwodrp);
      }
  printf("File Delays with drops in file_delays_drp.txt\n");
  printf("File Delays without drops in file_delays_ndrp.txt\n");
  fprintf(fp1,"File Delays with drops in file_delays_drp.txt\n");
  fprintf(fp1,"File Delays without drops in file_delays_ndrp.txt\n");
  fprintf(fp1,"\n");
  fprintf(fp1,"Number of files with drops:    %ld\n", filedropcnt);
  fprintf(fp1,"Total packets dropped:     %ld\n", dropcnt + dropcnt1);
  fprintf(fp1,"Total percent packet loss for test:  %f\n", 
	 ((float)(dropcnt + dropcnt1)/(float)(valcnt + valcnt1))*100.0);
  fprintf(fp1,"Mean file latency with drops:        %f\n", mean);
  fprintf(fp1,"Variance latency with drops:         %f\n", varval/filedropcnt);
  fprintf(fp1,"Standard Dev latency with drops:     %f\n", sqrt(varval/filedropcnt));
  fprintf(fp1,"Mean file latency without drops:     %f\n", mwodrp);
  fprintf(fp1,"Standard Dev latency without drops:  %f\n", sqrt(varval1/cntwodrps));
  fprintf(fp1,"\n");
  fprintf(fp1,"Packet Drop stats FROM REMOTE\n");
  fprintf(fp1,"Number of packets dropped:     %ld\n", dropcnt);
  fprintf(fp1,"Percent packet loss for test:  %f\n", ((float)dropcnt/(float)valcnt)*100.0);
  fprintf(fp1,"Maximum drops per file:        %f\n", maxt);
  fprintf(fp1,"\n");
  fprintf(fp1,"Packet Drop stats TO REMOTE\n");
  fprintf(fp1,"Number of packets dropped:     %ld\n", dropcnt1);
  fprintf(fp1,"Percent packet loss for test:  %f\n", ((float)dropcnt1/(float)valcnt1)*100.0);
  fprintf(fp1,"Maximum drops per file:        %f\n", maxt1);

  fclose(fp1);
  fclose(fp3);
  fclose(fp4);

  /* Now print out the critical path graphs and calculate some of the critical */
  /* path aggregate information.  This includes the average time spent in each */
  /* one of the five states defined for critical path:  1) network, 2) server  */
  /* 3) client, 4) fast rtran, 5) timeout. The graphs should simply be straight*/
  /* lines with contributions from each type of delay to overall delay.        */
  /* The crit path time lines must be generated carefully.  We want to generate*/
  /* a file which splits out the points by delay type.  This requires us to    */
  /* sort the tline structure using delay type as the key.                     */
  /* Also print out the critical path in time line format - this makes analysis*/
  /* a bit easier (these are the .cptl files - the .cp files simply print out  */
  /* a single line for cp).                                                    */

  qsort(cp,cur_cp,sizeof(cp[0]),compare5);  /* Sort by delay type              */

  if(GRAPHTOOL == 0) /* GRAPHTOOL = 0 means generate files for gnuplot         */
    {
      if((fp3 = fopen("net.cp", "w")) == NULL)
	{
	  perror("Unable to open net.cp\n");
	  exit(1);
	}
      if((fp4 = fopen("srv.cp", "w")) == NULL)
	{
	  perror("Unable to open srv.cp\n");
	  exit(1);
	}
      if((fp5 = fopen("cli.cp", "w")) == NULL)
	{
	  perror("Unable to open cli.cp\n");
	  exit(1);
	}
      if((fp6 = fopen("drp.cp", "w")) == NULL)
	{
	  perror("Unable to open drp.cp\n");
	  exit(1);
	}
      if((fp7 = fopen("tmo.cp", "w")) == NULL)
	{
	  perror("Unable to open tmo.cp\n");
	  exit(1);
	}
      if((fp8 = fopen("net.cptl", "w")) == NULL)
	{
	  perror("Unable to open net.cptl\n");
	  exit(1);
	}
      if((fp9 = fopen("srv.cptl", "w")) == NULL)
	{
	  perror("Unable to open srv.cptl\n");
	  exit(1);
	}
      if((fp10 = fopen("cli.cptl", "w")) == NULL)
	{
	  perror("Unable to open cli.cptl\n");
	  exit(1);
	}
      if((fp11 = fopen("drp.cptl", "w")) == NULL)
	{
	  perror("Unable to open drp.cptl\n");
	  exit(1);
	}
      if((fp12 = fopen("tmo.cptl", "w")) == NULL)
	{
	  perror("Unable to open tmo.cptl\n");
	  exit(1);
	}
      for(i=0;i<cur_cp;i++)
	{
	  switch(cp[i].dlytype) {
	  case 1:
	    fprintf(fp3,"%f %f\n", cp[i].x1, cp[i].y1);
	    fprintf(fp3,"%f %f\n", cp[i].x1, cp[i].y2);
	    fprintf(fp3,"\n");
	    if(cp[i].src) { /* if src = 1 source was client                    */
	      fprintf(fp8,"%f %f\n", cp[i].x1 - 1.5, cp[i].y1);
	      fprintf(fp8,"%f %f\n", cp[i].x1 - 0.5, cp[i].y2);
	    } else {        /* otherwise the source was the server             */
	      fprintf(fp8,"%f %f\n", cp[i].x1 - 0.5, cp[i].y1);
	      fprintf(fp8,"%f %f\n", cp[i].x1 - 1.5, cp[i].y2);
	    }
	    fprintf(fp8,"\n");
	    break;
	  case 2:
	    fprintf(fp4,"%f %f\n", cp[i].x1, cp[i].y1);
	    fprintf(fp4,"%f %f\n", cp[i].x1, cp[i].y2);
	    fprintf(fp4,"\n");
	    fprintf(fp9,"%f %f\n", cp[i].x1 - 0.5, cp[i].y1);
	    fprintf(fp9,"%f %f\n", cp[i].x1 - 0.5, cp[i].y2);
	    fprintf(fp9,"\n");
	    break;
	  case 3:
	    fprintf(fp5,"%f %f\n", cp[i].x1, cp[i].y1);
	    fprintf(fp5,"%f %f\n", cp[i].x1, cp[i].y2);
	    fprintf(fp5,"\n");
	    fprintf(fp10,"%f %f\n", cp[i].x1 - 1.5, cp[i].y1);
	    fprintf(fp10,"%f %f\n", cp[i].x1 - 1.5, cp[i].y2);
	    fprintf(fp10,"\n");
	    break;
	  case 4:
	    fprintf(fp6,"%f %f\n", cp[i].x1, cp[i].y1);
	    fprintf(fp6,"%f %f\n", cp[i].x1, cp[i].y2);
	    fprintf(fp6,"\n");
	    if(cp[i].src) { /* if src = 1 source was client                    */
	      fprintf(fp11,"%f %f\n", cp[i].x1 - 1.5, cp[i].y1);
	      fprintf(fp11,"%f %f\n", cp[i].x1 - 1.5, cp[i].y2);
	    } else {        /* otherwise the source was the server             */
	      fprintf(fp11,"%f %f\n", cp[i].x1 - 0.5, cp[i].y1);
	      fprintf(fp11,"%f %f\n", cp[i].x1 - 0.5, cp[i].y2);
	    }
	    fprintf(fp11,"\n"); /* print straight line for each drop too       */
	    fprintf(fp11,"%f %f\n", cp[i].x1 - 1.5, cp[i].y1);
	    fprintf(fp11,"%f %f\n", cp[i].x1 - 0.5, cp[i].y1);
	    fprintf(fp11,"\n");
	    break;
	  case 5:
	    fprintf(fp7,"%f %f\n", cp[i].x1, cp[i].y1);
	    fprintf(fp7,"%f %f\n", cp[i].x1, cp[i].y2);
	    fprintf(fp7,"\n");
	    if(cp[i].src) { /* if src = 1 source was client                    */
	      fprintf(fp12,"%f %f\n", cp[i].x1 - 1.5, cp[i].y1);
	      fprintf(fp12,"%f %f\n", cp[i].x1 - 1.5, cp[i].y2);
	    } else {        /* otherwise the source was the server             */
	      fprintf(fp12,"%f %f\n", cp[i].x1 - 0.5, cp[i].y1);
	      fprintf(fp12,"%f %f\n", cp[i].x1 - 0.5, cp[i].y2);
	    }
	    fprintf(fp12,"\n"); /* print straight line for each drop too       */
	    fprintf(fp12,"%f %f\n", cp[i].x1 - 1.5, cp[i].y1);
	    fprintf(fp12,"%f %f\n", cp[i].x1 - 0.5, cp[i].y1);
	    fprintf(fp12,"\n");
	    break;
	  }
	}
      fclose(fp3);
      fclose(fp4);
      fclose(fp5);
      fclose(fp6);
      fclose(fp7);
      fclose(fp8);
      fclose(fp9);
      fclose(fp10);
      fclose(fp11);
      fclose(fp12);

    } else {     /* GRAPHTOOL = 1 means generate files for xgraph              */

      if((fp3 = fopen("xgraph.cp", "w")) == NULL)
	{
	  perror("Unable to open xgraph.cp\n");
	  exit(1);
	}
      if((fp4 = fopen("xgraph.cptl", "w")) == NULL)
	{
	  perror("Unable to open xgraph.cptl\n");
	  exit(1);
	}
      seccnt = 0;
      for(i=0;i<cur_cp;i++)
	{
	  switch(cp[i].dlytype) {
	  case 1:
	    if(seccnt == 0)
	      {
		fprintf(fp3,"\"NETWORK\"\n");
		fprintf(fp4,"\"NETWORK\"\n");
		seccnt++;
	      }
	    fprintf(fp3,"move %f %f\n", cp[i].x1, cp[i].y1);
	    fprintf(fp3,"draw %f %f\n", cp[i].x1, cp[i].y2);
	    if(cp[i].src) { /* if src = 1 source was client                    */
	      fprintf(fp4,"move %f %f\n", cp[i].x1 - 1.5, cp[i].y1);
	      fprintf(fp4,"draw %f %f\n", cp[i].x1 - 0.5, cp[i].y2);
	    } else {        /* otherwise the source was the server             */
	      fprintf(fp4,"move %f %f\n", cp[i].x1 - 0.5, cp[i].y1);
	      fprintf(fp4,"draw %f %f\n", cp[i].x1 - 1.5, cp[i].y2);
	    }
	    break;
	  case 2:
	    if(seccnt == 1)
	      {
		fprintf(fp3,"\n\"SERVER\"\n");
		fprintf(fp4,"\n\"SERVER\"\n");
		seccnt++;
	      }
	    fprintf(fp3,"move %f %f\n", cp[i].x1, cp[i].y1);
	    fprintf(fp3,"draw %f %f\n", cp[i].x1, cp[i].y2);
	    fprintf(fp4,"move %f %f\n", cp[i].x1 - 0.5, cp[i].y1);
	    fprintf(fp4,"draw %f %f\n", cp[i].x1 - 0.5, cp[i].y2);
	    break;
	  case 3:
	    if(seccnt <= 2)
	      {
		fprintf(fp3,"\n\"CLIENT\"\n");
		fprintf(fp4,"\n\"CLIENT\"\n");
		seccnt = 3;
	      }
	    fprintf(fp3,"move %f %f\n", cp[i].x1, cp[i].y1);
	    fprintf(fp3,"draw %f %f\n", cp[i].x1, cp[i].y2);
	    fprintf(fp4,"move %f %f\n", cp[i].x1 - 1.5, cp[i].y1);
	    fprintf(fp4,"draw %f %f\n", cp[i].x1 - 1.5, cp[i].y2);
	    break;
	  case 4:
	    if(seccnt <= 3)
	      {
		fprintf(fp3,"\n\"FAST RETRAN\"\n");
		fprintf(fp4,"\n\"FAST RETRAN\"\n");
		seccnt = 4;
	      }
	    fprintf(fp3,"move %f %f\n", cp[i].x1, cp[i].y1);
	    fprintf(fp3,"draw %f %f\n", cp[i].x1, cp[i].y2);
	    if(cp[i].src) { /* if src = 1 source was client                    */
	      fprintf(fp4,"move %f %f\n", cp[i].x1 - 1.5, cp[i].y1);
	      fprintf(fp4,"draw %f %f\n", cp[i].x1 - 1.5, cp[i].y2);
	    } else {        /* otherwise the source was the server             */
	      fprintf(fp4,"move %f %f\n", cp[i].x1 - 0.5, cp[i].y1);
	      fprintf(fp4,"draw %f %f\n", cp[i].x1 - 0.5, cp[i].y2);
	    }  /* print straight line where the drop took place                */
	    fprintf(fp4,"move %f %f\n", cp[i].x1 - 1.5, cp[i].y1);
	    fprintf(fp4,"draw %f %f\n", cp[i].x1 - 0.5, cp[i].y1);
	    break;
	  case 5:
	    if(seccnt <= 4)
	      {
		fprintf(fp3,"\n\"TIME OUT\"\n");
		fprintf(fp4,"\n\"TIME OUT\"\n");
		seccnt = 5;
	      }
	    fprintf(fp3,"move %f %f\n", cp[i].x1, cp[i].y1);
	    fprintf(fp3,"draw %f %f\n", cp[i].x1, cp[i].y2);
	    if(cp[i].src) { /* if src = 1 source was client                    */
	      fprintf(fp4,"move %f %f\n", cp[i].x1 - 1.5, cp[i].y1);
	      fprintf(fp4,"draw %f %f\n", cp[i].x1 - 1.5, cp[i].y2);
	    } else {        /* otherwise the source was the server             */
	      fprintf(fp4,"move %f %f\n", cp[i].x1 - 0.5, cp[i].y1);
	      fprintf(fp4,"draw %f %f\n", cp[i].x1 - 0.5, cp[i].y2);
	    }  /* print straight line where the drop took place                */
	    fprintf(fp4,"move %f %f\n", cp[i].x1 - 1.5, cp[i].y1);
	    fprintf(fp4,"draw %f %f\n", cp[i].x1 - 0.5, cp[i].y1);
	    break;
	  }
	}
      fclose(fp3);
      fclose(fp4);
    }

  /* Now print out the files which contain the critical path delay info for    */
  /* each transfer.  These can be used in a statistical analysis of delays.    */
  /* Print both the delays each individual component and aggregate for each    */
  /* file transferred.  Omit any transfer which has the abort flag set.        */
  /* In addition to the data also generate mean and variance for each type of  */
  /* delay.                                                                    */
  /* In addition to cp information, also print out the connection setup,       */
  /* GET/transfer and connection tear down delays for each transfer.  Also     */
  /* print out the mean and variance for each of these as well.                */

  qsort(cp,cur_cp,sizeof(cp[0]),compare9);  /* Sort by file for per file cp    */

  if((fp3 = fopen("cpnet.dly", "w")) == NULL)
    {
      perror("Unable to open cpnet.dly\n");
      exit(1);
    }
  if((fp4 = fopen("cpsrv.dly", "w")) == NULL)
    {
      perror("Unable to open cpsrv.dly\n");
      exit(1);
    }
  if((fp5 = fopen("cpcli.dly", "w")) == NULL)
    {
      perror("Unable to open cpcli.dly\n");
      exit(1);
    }
  if((fp6 = fopen("cpdrp.dly", "w")) == NULL)
    {
      perror("Unable to open cpdrp.dly\n");
      exit(1);
    }
  if((fp7 = fopen("cptmo.dly", "w")) == NULL)
    {
      perror("Unable to open cptmo.dly\n");
      exit(1);
    }
  if((fp8 = fopen("cpnetfile.dly", "w")) == NULL)
    {
      perror("Unable to open cpnetfile.dly\n");
      exit(1);
    }
  if((fp9 = fopen("cpsrvfile.dly", "w")) == NULL)
    {
      perror("Unable to open cpsrvfile.dly\n");
      exit(1);
    }
  if((fp10 = fopen("cpclifile.dly", "w")) == NULL)
    {
      perror("Unable to open cpclifile.dly\n");
      exit(1);
    }
  if((fp11 = fopen("cpdrpfile.dly", "w")) == NULL)
    {
      perror("Unable to open cpdrpfile.dly\n");
      exit(1);
    }
  if((fp12 = fopen("cptmofile.dly", "w")) == NULL)
    {
      perror("Unable to open cptmofile.dly\n");
      exit(1);
    }
  if((fp13 = fopen("cp_summary.txt", "w")) == NULL)
    {
      perror("Unable to open cp_summary.txt\n");
      exit(1);
    }
  if((fp14 = fopen("consetup.dly", "w")) == NULL)
    {
      perror("Unable to open consetup.dly\n");
      exit(1);
    }
  if((fp15 = fopen("contransfer.dly", "w")) == NULL)
    {
      perror("Unable to open transfer.dly\n");
      exit(1);
    }
  if((fp16 = fopen("conteardown.dly", "w")) == NULL)
    {
      perror("Unable to open conteardown.dly\n");
      exit(1);
    }
  if((fp17 = fopen("cp_descript.txt", "w")) == NULL)
    {
      perror("Unable to open cp_descript.txt\n");
      exit(1);
    }
  mean = varval = 0.0;        /* Mean and variance for cp network delays       */
  mean1 = varval1 = 0.0;      /* Mean and variance for cp server delays        */
  mean2 = varval2 = 0.0;      /* Mean and variance for cp client delays        */
  mean3 = varval3 = 0.0;      /* Mean and variance for cp drop delays          */
  mean4 = varval4 = 0.0;      /* Mean and variance for cp time out delays      */
  mean5 = varval5 = 0.0;      /* Mean and variance for connection setup delays */
  mean6 = varval6 = 0.0;      /* Mean and variance for GET/transfer delays     */
  mean7 = varval7 = 0.0;      /* Mean and variance for connect tear down delays*/
  mean8 = varval8 = 0.0;      /* Mean and variance for cp file network delays  */
  mean9 = varval9 = 0.0;      /* Mean and variance for cp file server delays   */
  mean10 = varval10 = 0.0;    /* Mean and variance for cp file client delays   */
  mean11 = varval11 = 0.0;    /* Mean and variance for cp file drop delays     */
  mean12 = varval12 = 0.0;    /* Mean and variance for cp file time out delays */
  mean13 = varval13 = 0.0;    /* Mean and variance for # file time out delays  */
  mean14 = varval14 = 0.0;    /* Mean and variance for # file fast rt delays   */
  cpcnt = cpcnt1 = cpcnt2 = cpcnt3 = cpcnt4 = cpcnt5 = 0;  /* Instance counters*/
  maxt = maxt1 = 0;    /* Used to track setup, GET/transfer, tear down         */
  mint2 = mint;        /* Clnt - Srv min prop. delay = mint2,Srv - Clnt = mint1*/
  mint = 10.0;         /* False min time just to initialize                    */
  cpcli = cpsrv = cpnet = cptmo = cpdrp = 0.0;  /* Means for cp file delays    */
  startflag = 0;     /* Use this to track if there was an abort for a transfer */
  valcnt = 1;
  for(i=0;i<cur_cp;i++)
    {
      if(cp[i].abort == 0) {   /* Only consider transfers which weren't aborted*/
	if(valcnt != cp[i].file) {
	  fprintf(fp14,"%f\n", mint);
	  fprintf(fp15,"%f\n", maxt - mint);
	  fprintf(fp16,"%f\n", maxt1 - maxt);
	  mean5 += mint;
	  mean6 += maxt - mint;
	  mean7 += maxt1 - maxt;
	  mean8 += cpnet;
	  mean9 += cpsrv;
	  mean10 += cpcli;
	  mean11 += cpdrp;
	  mean12 += cptmo;
	  cpcnt5++;
	  cpcli = cpsrv = cpnet = cptmo = cpdrp = 0.0;
	  valcnt = cp[i].file;
	  maxt = maxt1 = 0.0;
	  mint = 10.0;
	  startflag = 0;
	}
	switch(cp[i].status) { /*Get delay info for setup, transfer or teardown*/
	case 2:            /* Get min and max times for GET/Data transfer  */
	  if( cp[i].y2 > cp[i].y1) {
	    if(cp[i].y2 > maxt) maxt = cp[i].y2;
	    if(cp[i].y1 < mint) mint = cp[i].y1;
	  } else {
	    if(cp[i].y1 > maxt) maxt = cp[i].y1;
	    if(cp[i].y2 < mint) mint = cp[i].y2;
	  }
	  break;
	case 3:            /* Get max time for connection tear down        */
	  if( cp[i].y2 > cp[i].y1) {
	    if(cp[i].y2 > maxt1) maxt1 = cp[i].y2;
	  } else {
	    if(cp[i].y1 > maxt1) maxt1 = cp[i].y1;
	  }
	  break;
	}
	switch(cp[i].dlytype) {  /* Pick out the individual delay types on cp  */
	case 1:
	  if( cp[i].y2 > cp[i].y1) {
	    fprintf(fp3,"%f\n", cp[i].y2 - cp[i].y1);
	    mean += cp[i].y2 - cp[i].y1;
	    cpnet += cp[i].y2 - cp[i].y1;
	    fprintf(fp17,"Rport: %ld, Start: %f, Net delay: %f, IPID %ld\n",
		    cp[i].rport,cp[i].y1,cp[i].y2 - cp[i].y1,cp[i].ipid);
	  } else {
	    fprintf(fp3,"%f\n", cp[i].y1 - cp[i].y2);
	    mean += cp[i].y1 - cp[i].y2;
	    cpnet += cp[i].y1 - cp[i].y2;
	    fprintf(fp17,"Rport: %ld, Start: %f, Net delay: %f, IPID %ld\n",
		    cp[i].rport,cp[i].y2,cp[i].y1 - cp[i].y2,cp[i].ipid);
	  }
	  cpcnt++;
	  break;
	case 2:
	  if( cp[i].y2 > cp[i].y1) {
	    fprintf(fp4,"%f\n", cp[i].y2 - cp[i].y1);
	    mean1 += cp[i].y2 - cp[i].y1;
	    cpsrv += cp[i].y2 - cp[i].y1;
	    fprintf(fp17,"Rport: %ld, Start: %f, Srv delay: %f\n",
		    cp[i].rport,cp[i].y1,cp[i].y2 - cp[i].y1);
	  } else {
	    fprintf(fp4,"%f\n", cp[i].y1 - cp[i].y2);
	    mean1 += cp[i].y1 - cp[i].y2;
	    cpsrv += cp[i].y1 - cp[i].y2;
	    fprintf(fp17,"Rport: %ld, Start: %f, Srv delay: %f\n",
		    cp[i].rport,cp[i].y2,cp[i].y1 - cp[i].y2);
	  }
	  cpcnt1++;
	  break;
	case 3:
	  if( cp[i].y2 > cp[i].y1) {
	    fprintf(fp5,"%f\n", cp[i].y2 - cp[i].y1);
	    mean2 += cp[i].y2 - cp[i].y1;
	    cpcli += cp[i].y2 - cp[i].y1;
	    fprintf(fp17,"Rport: %ld, Start: %f, Cli delay: %f\n",
		    cp[i].rport,cp[i].y1,cp[i].y2 - cp[i].y1);
	  } else {
	    fprintf(fp5,"%f\n", cp[i].y1 - cp[i].y2);
	    mean2 += cp[i].y1 - cp[i].y2;
	    cpcli += cp[i].y1 - cp[i].y2;
	    fprintf(fp17,"Rport: %ld, Start: %f, Srv delay: %f\n",
		    cp[i].rport,cp[i].y2, cp[i].y1 - cp[i].y2);
	  }
	  cpcnt2++;
	  break;
	case 4:
	  if( cp[i].y2 > cp[i].y1) {
	    fprintf(fp6,"%f\n", cp[i].y2 - cp[i].y1);
	    mean3 += cp[i].y2 - cp[i].y1;
	    cpdrp += cp[i].y2 - cp[i].y1;
	    fprintf(fp17,"Rport: %ld, Start: %f, Fst Rtrn delay: %f, IPID %ld\n",
		    cp[i].rport,cp[i].y1,cp[i].y2 - cp[i].y1,cp[i].ipid);
	  } else {
	    fprintf(fp6,"%f\n", cp[i].y1 - cp[i].y2);
	    mean3 += cp[i].y1 - cp[i].y2;
	    cpcli += cp[i].y1 - cp[i].y2;
	    fprintf(fp17,"Rport: %ld, Start: %f, Fst Rtrn delay: %f, IPID %ld\n",
		    cp[i].rport,cp[i].y2,cp[i].y1 - cp[i].y2,cp[i].ipid);
	  }
	  cpcnt3++;
	  break;
	case 5:
	  if( cp[i].y2 > cp[i].y1) {
	    fprintf(fp7,"%f\n", cp[i].y2 - cp[i].y1);
	    mean4 += cp[i].y2 - cp[i].y1;
	    cptmo += cp[i].y2 - cp[i].y1;
	    fprintf(fp17,"Rport: %ld, Start: %f, Tmo delay: %f, IPID %ld\n",
		    cp[i].rport,cp[i].y1,cp[i].y2 - cp[i].y1,cp[i].ipid);
	  } else {
	    fprintf(fp7,"%f\n", cp[i].y1 - cp[i].y2);
	    mean4 += cp[i].y1 - cp[i].y2;
	    cptmo += cp[i].y1 - cp[i].y2;
	    fprintf(fp17,"Rport: %ld, Start: %f, Tmo delay: %f, IPID %ld\n",
		    cp[i].rport,cp[i].y2,cp[i].y1 - cp[i].y2,cp[i].ipid);
	  }
	  cpcnt4++;
	  break;
	}
      } else {            /* Set startflag = 1 if an abort is found in the transfer */
	startflag = 1;
      }
      if(i == cur_cp - 1 && startflag == 0) {  /* Print out the last transfer       */
	fprintf(fp14,"%f\n", mint);
	fprintf(fp15,"%f\n", maxt - mint);
	fprintf(fp16,"%f\n", maxt1 - maxt);
	mean5 += mint;
	mean6 += maxt - mint;
	mean7 += maxt1 - maxt;
	mean8 += cpnet;
	mean9 += cpsrv;
	mean10 += cpcli;
	mean11 += cpdrp;
	mean12 += cptmo;
	cpcnt5++;
      }
    }   /* End for loop       */
  if(cpcnt > 0) mean = mean/cpcnt;
  if(cpcnt1 > 0) mean1 = mean1/cpcnt1;
  if(cpcnt2 > 0) mean2 = mean2/cpcnt2;
  if(cpcnt3 > 0) mean3 = mean3/cpcnt3;
  if(cpcnt4 > 0) mean4 = mean4/cpcnt4;
  if(cpcnt5 > 0) {
    mean5 = mean5/cpcnt5;
    mean6 = mean6/cpcnt5;
    mean7 = mean7/cpcnt5;
    mean8 = mean8/cpcnt5;
    mean9 = mean9/cpcnt5;
    mean10 = mean10/cpcnt5;
    mean11 = mean11/cpcnt5;
    mean12 = mean12/cpcnt5;
    mean13 = (float)cpcnt3/cpcnt5;
    mean14 = (float)cpcnt4/cpcnt5;
  }

  /* Now calculate variance for individual and file cp stats.                     */

  cpcli = 0.0;
  cpsrv = 0.0;
  cpnet = 0.0;
  cptmo = 0.0;
  cpdrp = 0.0;
  cpcnt6 = 0;
  cpcnt7 = 0;
  cpcnt8 = 0;
  cpcnt9 = 0;
  valcnt = 1;
  startflag = 0;        /* Use this to track if there was an abort for a transfer */
  j = 0;                /* Use this to flag when to end tear down variance calc   */
  for(i=0;i<cur_cp;i++)
    {
      if(cp[i].abort == 0) {
	if(valcnt != cp[i].file) {
	  if(startflag == 0) {
	    fprintf(fp8,"%f\n", cpnet);
	    fprintf(fp9,"%f\n", cpsrv);
	    fprintf(fp10,"%f\n", cpcli);
	    fprintf(fp11,"%f\n", cpdrp);
	    fprintf(fp12,"%f\n", cptmo);
	  }
	  varval5 += (mint - mean5) * (mint - mean5);
	  varval6 += ((maxt - mint) - mean6) * ((maxt - mint) - mean6);
	  varval7 += ((maxt1 - maxt) - mean7) * ((maxt1 - maxt) - mean7);
	  varval8 += (cpnet - mean8) * (cpnet - mean8);
	  varval9 += (cpsrv - mean9) * (cpsrv - mean9);
	  varval10 += (cpcli - mean10) * (cpcli - mean10);
	  varval11 += (cpdrp - mean11) * (cpdrp - mean11);
	  varval12 += (cptmo - mean12) * (cptmo - mean12);
	  varval13 += (cpcnt6 - mean13) * (cpcnt6 - mean13);
	  varval14 += (cpcnt7 - mean14) * (cpcnt7 - mean14);
	  cpcli = 0.0;
	  cpsrv = 0.0;
	  cpnet = 0.0;
	  cptmo = 0.0;
	  cpdrp = 0.0;
	  cpcnt6 = 0;
	  cpcnt7 = 0;
	  maxt = maxt1 = 0.0;
	  mint = 10.0;
	  valcnt = cp[i].file;
	  j = i - 1;
	  startflag = 0;
	}
	switch(cp[i].status) {  /* Assign delay to setup, transfer or tear down*/
	case 2:            /* Assign delay to GET/Data transfer              */
	  if( cp[i].y2 > cp[i].y1) {
	    if(cp[i].y2 > maxt) maxt = cp[i].y2;
	    if(cp[i].y1 < mint) mint = cp[i].y1;
	  } else {
	    if(cp[i].y1 > maxt) maxt = cp[i].y1;
	    if(cp[i].y2 < mint) mint = cp[i].y2;
	  }
	  break;
	case 3:            /* Assign delay to connection tear down           */
	  if( cp[i].y2 > cp[i].y1) {
	    if(cp[i].y2 > maxt1) maxt1 = cp[i].y2;
	  } else {
	    if(cp[i].y1 > maxt1) maxt1 = cp[i].y1;
	  }
	  break;
	}
	switch(cp[i].dlytype) {
	case 1:
	  if( cp[i].y2 > cp[i].y1) {
	    cpnet += cp[i].y2 - cp[i].y1;
	    varval += ((cp[i].y2 - cp[i].y1) - mean) * ((cp[i].y2 - cp[i].y1) - mean);
	  } else {
	    cpnet += cp[i].y1 - cp[i].y2;
	    varval += ((cp[i].y1 - cp[i].y2) - mean) * ((cp[i].y1 - cp[i].y2) - mean);
	  }
	  if(cp[i].host == 1) {       /* If host = 1, source = client (HOST2) */
	    cpcnt8++;  /* Increment count of net delays from client to server */ 
	  } else {
	    cpcnt9++;  /* Increment count of net delays from server to client */ 
	  }
	  break;
	case 2:
	  if( cp[i].y2 > cp[i].y1) {
	    cpsrv += cp[i].y2 - cp[i].y1;
	    varval1 += ((cp[i].y2 - cp[i].y1) - mean1) * ((cp[i].y2 - cp[i].y1) - mean1);
	  } else {
	    cpsrv += cp[i].y1 - cp[i].y2;
	    varval1 += ((cp[i].y1 - cp[i].y2) - mean1) * ((cp[i].y1 - cp[i].y2) - mean1);
	  }
	  break;
	case 3:
	  if( cp[i].y2 > cp[i].y1) {
	    cpcli += cp[i].y2 - cp[i].y1;
	    varval2 += ((cp[i].y2 - cp[i].y1) - mean2) * ((cp[i].y2 - cp[i].y1) - mean2);
	  } else {
	    cpcli += cp[i].y1 - cp[i].y2;
	    varval2 += ((cp[i].y1 - cp[i].y2) - mean2) * ((cp[i].y1 - cp[i].y2) - mean2);
	  }
	  break;
	case 4:
	  if( cp[i].y2 > cp[i].y1) {
	    cpdrp += cp[i].y2 - cp[i].y1;
	    varval3 += ((cp[i].y2 - cp[i].y1) - mean3) * ((cp[i].y2 - cp[i].y1) - mean3);
	  } else {
	    cpdrp += cp[i].y1 - cp[i].y2;
	    varval3 += ((cp[i].y1 - cp[i].y2) - mean3) * ((cp[i].y1 - cp[i].y2) - mean3);
	  }
	  cpcnt6++;
	  break;
	case 5:
	  if( cp[i].y2 > cp[i].y1) {
	    cptmo += cp[i].y2 - cp[i].y1;
	    varval4 += ((cp[i].y2 - cp[i].y1) - mean4) * ((cp[i].y2 - cp[i].y1) - mean4);
	  } else {
	    cptmo += cp[i].y1 - cp[i].y2;
	    varval4 += ((cp[i].y1 - cp[i].y2) - mean4) * ((cp[i].y1 - cp[i].y2) - mean4);
	  }
	  cpcnt7++;
	  break;
	}
      } else {
	startflag = 1;  /* Set startflag = 1 if an abort is found in the transfer */
      }
      if(i == cur_cp - 1 && startflag == 0) {  /* Print out the last transfer     */
	fprintf(fp8,"%f\n", cpnet);
	fprintf(fp9,"%f\n", cpsrv);
	fprintf(fp10,"%f\n", cpcli);
	fprintf(fp11,"%f\n", cpdrp);
	fprintf(fp12,"%f\n", cptmo);
	varval5 += (mint - mean5) * (mint - mean5);
	varval6 += ((maxt - mint) - mean6) * ((maxt - mint) - mean6);
	varval7 += ((maxt1 - maxt) - mean7) * ((maxt1 - maxt) - mean7);
	varval13 += (cpcnt6 - mean13) * (cpcnt6 - mean13);
	varval14 += (cpcnt7 - mean14) * (cpcnt7 - mean14);
      }
    }

  printf("Critical path summary statistics in file cp_summary.txt\n");
  fprintf(fp13,"Mean per packet network latency:        %f\n", mean);
  fprintf(fp13,"Variance per packet network latency:         %f\n", varval/cpcnt);
  fprintf(fp13,"Std Dev per packet network latency:     %f\n", sqrt(varval/cpcnt));
  fprintf(fp13,"Min packet propagation delay Server - Client:           %f\n",mint1);
  fprintf(fp13,"Min packet propagation delay Client - Server:           %f\n",mint2);
  fprintf(fp13,"Average per packet queuing delay for cp:        %f\n\n",mean - (mint1+mint2)/2);
  fprintf(fp13,"Mean per instance server latency:        %f\n", mean1);
  fprintf(fp13,"Variance per instance server latency:         %f\n", varval1/cpcnt1);
  fprintf(fp13,"Std Dev per instance server latency:     %f\n\n", sqrt(varval1/cpcnt1));
  fprintf(fp13,"Mean per instance client latency:        %f\n", mean2);
  fprintf(fp13,"Variance per instance client latency:         %f\n", varval2/cpcnt2);
  fprintf(fp13,"Std Dev per instance client latency:     %f\n\n", sqrt(varval2/cpcnt2));
  fprintf(fp13,"Mean per instance drop latency:        %f\n", mean3);
  fprintf(fp13,"Variance per instance drop latency:         %f\n", varval3/cpcnt3);
  fprintf(fp13,"Std Dev per instance drop latency:     %f\n\n", sqrt(varval3/cpcnt3));
  fprintf(fp13,"Mean per instance time-out latency:        %f\n", mean4);
  fprintf(fp13,"Variance per instance timeout latency:         %f\n", varval4/cpcnt4);
  fprintf(fp13,"Std Dev per instance timeout latency:     %f\n\n", sqrt(varval4/cpcnt4));
  fprintf(fp13,"Mean per file network latency:        %f\n", mean8);
  fprintf(fp13,"Variance per file network latency:         %f\n", varval8/cpcnt5);
  fprintf(fp13,"Std Dev per file network latency:     %f\n", sqrt(varval8/cpcnt5));
  fprintf(fp13,"Minimum per file propagation delay:           %f\n",
	  (mint1 * cpcnt9/cpcnt5) + (mint2 * cpcnt8/cpcnt5));
  fprintf(fp13,"Average per file queuing delay for cp:  %f\n\n",
	  mean8 - (((mint1 + mint2)/2) * cpcnt/cpcnt5));
  fprintf(fp13,"Mean per file server latency:        %f\n", mean9);
  fprintf(fp13,"Variance per file server latency:         %f\n", varval9/cpcnt5);
  fprintf(fp13,"Std Dev per file server latency:     %f\n\n", sqrt(varval9/cpcnt5));
  fprintf(fp13,"Mean per file client latency:        %f\n", mean10);
  fprintf(fp13,"Variance per file client latency:         %f\n", varval10/cpcnt5);
  fprintf(fp13,"Std Dev per file client latency:     %f\n\n", sqrt(varval10/cpcnt5));
  fprintf(fp13,"Mean number drops per file:        %f\n", mean13);
  fprintf(fp13,"Variance number drops per file:         %f\n", varval13/cpcnt5);
  fprintf(fp13,"Mean per file drop latency:        %f\n", mean11);
  fprintf(fp13,"Variance per file drop latency:         %f\n", varval11/cpcnt5);
  fprintf(fp13,"Std Dev per file drop latency:     %f\n\n", sqrt(varval11/cpcnt5));
  fprintf(fp13,"Mean number time-outs per file:        %f\n", mean14);
  fprintf(fp13,"Variance number time-outs per file:         %f\n", varval14/cpcnt5);
  fprintf(fp13,"Mean per file time-out latency:        %f\n", mean12);
  fprintf(fp13,"Variance per file timeout latency:         %f\n", varval12/cpcnt5);
  fprintf(fp13,"Std Dev per file timeout latency:     %f\n\n", sqrt(varval12/cpcnt5));
  fprintf(fp13,"Mean connection setup latency:        %f\n", mean5);
  fprintf(fp13,"Variance connection setup latency:         %f\n", varval5/cpcnt5);
  fprintf(fp13,"Standard Dev connection setup latency:     %f\n\n", sqrt(varval5/cpcnt5));
  fprintf(fp13,"Mean GET/data transfer latency:        %f\n", mean6);
  fprintf(fp13,"Variance GET/data transfer latency:         %f\n", varval6/cpcnt5);
  fprintf(fp13,"Standard Dev GET/data transfer latency:     %f\n\n", sqrt(varval6/cpcnt5));
  fprintf(fp13,"Mean connection tear down latency:        %f\n", mean7);
  fprintf(fp13,"Variance connection tear down latency:         %f\n", varval7/cpcnt5);
  fprintf(fp13,"Std Dev connection tear down latency:     %f\n", sqrt(varval7/cpcnt5));

  fclose(fp3);
  fclose(fp4);
  fclose(fp5);
  fclose(fp6);
  fclose(fp7);
  fclose(fp8);
  fclose(fp9);
  fclose(fp10);
  fclose(fp11);
  fclose(fp12);
  fclose(fp13);
  fclose(fp14);
  fclose(fp15);
  fclose(fp16);
  fclose(fp17);
  
  /* Now print out the simple TCP packet and byte count information.  This can */
  /* be used in a variety of ways including time variance plots for self       */
  /* similar behavior as well as just visualizing the generat transfer behavior*/

  if((fp3 = fopen("tcpcounts.txt", "w")) == NULL)
    {
      perror("Unable to open tcpcount.txt\n");
      exit(1);
    }
  for(i=0;i<=num_intv;i++)
    {
      fprintf(fp3,"%ld\n", pktcount[i]);
    }
  fclose(fp3);
  
  if((fp3 = fopen("tcpbytes.txt", "w")) == NULL)
    {
      perror("Unable to open tcpbytes.txt\n");
      exit(1);
    }
  for(i=0;i<=num_intv;i++)
    {
      fprintf(fp3,"%ld\n", bytcount[i]);
    }
  fclose(fp3);


  /* Some people like to look at sequence number charts since they can show */
  /* some things that the timelines cannot.  The following section prints   */
  /* the packet sequence numbers sent and the ACK's which are returned as a */
  /* function of time.  This routine prints out one file for xgraph and two */
  /* for gnuplot which track data packets and ACKs. The graphs are printed  */
  /* such that all sequences start at 0 and values on the x-axis all begin  */
  /* on the next integer.  NOTE:  this only tracks the connection from the  */
  /* server to the client ie. the ACK's come from the client and data comes */
  /* from the server.  Sequence plots start after SYN sequence              */
  /* Begin by sorting the data by rport which should sort them into file xfrs*/

  qsort(outl,num_tln,sizeof(outl[0]),compare1);


  if(GRAPHTOOL == 0)  /* GRAPHTOOL = 0 means generate files for gnuplot     */
    {
      if((fp3 = fopen("data.seq", "w")) == NULL)
	{
	  perror("Unable to open data.seq\n");
	  exit(1);
	}
      if((fp4 = fopen("ack.seq", "w")) == NULL)
	{
	  perror("Unable to open ack.seq\n");
	  exit(1);
	}
      xmarker = 0.0;
      lastport = outl[0].rport;
      for(i=0;i<num_tln;i++)
	if(outl[i].pkttype <= 4 && lastport == outl[i].rport)
	  {
	    /* Track DATA from server                                       */
	    if(outl[i].host == 0 && outl[i].pkttype == 3) 
	       {
		   fprintf(fp3,"%f %ld\n", outl[i].y2 + xmarker, outl[i].seqn);
	       }
	    /* Track ACK's sent from client                                 */
	    if(outl[i].host == 1 && outl[i].pkttype == 2 && outl[i].seqn != 1)
	       {
		 fprintf(fp4,"%f %ld\n", outl[i].y2 + xmarker, outl[i].seqn + 1);
	       }
	  } else {
	    if(lastport != outl[i].rport) {
	      lastport = outl[i].rport; /* set current port                 */
	      xmarker = ceil(xmarker + outl[i - 1].y2);  /* round xmarket up*/
	      fprintf(fp3,"\n");
	      fprintf(fp4,"\n");
	      i--;  /* Be sure to include current packet in next sequence   */
	    }
	  }
      fclose(fp3);
      fclose(fp4);
    } else {          /* GRAPHTOOL = 1 means generate files for xgraph      */
      if((fp3 = fopen("xgraph.seq", "w")) == NULL)
	{
	  perror("Unable to open xgraph.seq\n");
	  exit(1);
	}
      startflag = 1;
      xmarker = 0.0;
      lastport = outl[0].rport;
      fprintf(fp3,"\"DATA FROM SERVER\"\n");
      for(i=0;i<num_tln;i++)  /* Track the data from the server first       */
	if(outl[i].pkttype <= 4 && lastport == outl[i].rport)
	  {
	    if(outl[i].host == 0 && outl[i].pkttype == 3)
	       {
		 if(startflag) {
		   fprintf(fp3,"move %f %ld\n", xmarker, outl[i].seqn);
		   startflag = 0;
		 } else {
		   fprintf(fp3,"draw %f %ld\n", outl[i].y2 + xmarker, outl[i].seqn);
		 }
	       }
	  } else {
	    if(lastport != outl[i].rport) {
	      lastport = outl[i].rport; /* set current port                 */
	      xmarker = ceil(xmarker + outl[i - 1].y2); /* round xmarket up */
	      i--;  /* Be sure to include current packet in next sequence   */
	      startflag = 1;
	    }
	  }
      startflag = 1;
      xmarker = 0.0;
      lastport = outl[0].rport;
      fprintf(fp3,"\n\"ACKS FROM CLIENT\"\n");
      for(i=0;i<num_tln;i++) /* Now track the acks from the client          */
	if(outl[i].pkttype <= 4 && lastport == outl[i].rport)
	  {
	    /* Track ACK's sent from client                                 */
	    if(outl[i].host == 1 && outl[i].pkttype == 2) 
	       {
		 if(startflag) {
		   fprintf(fp3,"move %f %ld\n", xmarker, outl[i].seqn + 1);
		   startflag = 0;
		 } else {
		   fprintf(fp3,"draw %f %ld\n", outl[i].y2 + xmarker, outl[i].seqn + 1);
		 }
	       }
	  } else {
	    if(lastport != outl[i].rport) {
	      lastport = outl[i].rport; /* set current port                 */
	      xmarker = ceil(xmarker + outl[i - 1].y2); /* round xmarket up */
	      i--;  /* Be sure to include current packet in next sequence   */
	      startflag = 1;
	    }
	  }
      fclose(fp3);
    }

  /* The following loop picks out the interpacket spacing for each file     */
  /* transferred.  This relies on the data array being sorted first by      */
  /* rport which was done above.  The loop finds the end of each transfer   */
  /* and then sorts the transfer by the time it left the server.  So, it    */
  /* only really tracks interpacket spacing times for data during the bulk  */
  /* transfer phase of a transaction.  The mean and varience are printed    */
  /* in intpkt_info.txt and the raw spacing times are in inkpkt_all.dly     */
  /* We also calculate the interpacket space both before and after a drop   */
  /* occurs in a transfer in order to understand both SS and Cong. avoid    */
  /* behaviors.  The before drop interpacket spacing is in inkpkt_ss.dly    */
  /* and the interpacket spacing after a drop is in inkpkt_cf.dlly.         */
  /* These loops also generate counts of packets which are sent out back-   */
  /* to-back.  Namely,, if a series of n packets all have the same time     */
  /* stamp at the server then the count of n gets entered into the count    */
  /* array.  Calculate the mean and variance of the back-to-back counts     */
  /* and print out a file of the counts.                                    */


  if((fp3 = fopen("intpkt_info.txt", "w")) == NULL)
    {
      perror("Unable to open intpkt_info.txt\n");
      exit(1);
    }
  if((fp4 = fopen("intpkt_all.dly", "w")) == NULL)
    {
      perror("Unable to open intpkt_all.dly\n");
      exit(1);
    }
  if((fp5 = fopen("intpkt_ss.dly", "w")) == NULL)
    {
      perror("Unable to open intpkt_all.dly\n");
      exit(1);
    }
  if((fp6 = fopen("intpkt_cf.dly", "w")) == NULL)
    {
      perror("Unable to open intpkt_all.dly\n");
      exit(1);
    }
  if((fp7 = fopen("back2back.cnt", "w")) == NULL)
    {
      perror("Unable to open back2back.cnt\n");
      exit(1);
    }
  
  mean = varval = 0.0;     /* Track mean and variance of all interpkt spacing */
  mean1 = varval1 = 0.0;   /* Track mean and variance of slow start spacing   */
  mean2 = varval2 = 0.0;   /* Track mean and variance of cong control spacing */
  mean3 = varval3 = 0.0;   /* Track mean and variance of back-to-back counts  */
  valcnt = 0;
  valcnt1 = 0;
  dropcnt = 0;
  backcnt = 0;             /* Count of back-to-back sequences in the data set */
  xmarker = 0.0;
  startflag = 0;
  startflag1 = 0;
  nodrp = 0;
  lastport = outl[0].rport;
  for(i=0;i<num_tln;i++)   /* First calcuate means and populate raw data files*/
    {                    
      if(lastport != outl[i].rport || i == num_tln - 1) /* Find file end point*/
	{
	  qsort(&outl[startflag],i - startflag,sizeof(outl[0]),compare7);

	  numback = 1;       /* Count of back to back packets                 */
	  for(j=startflag;j<i;j++)
	    {
	      if(outl[j].host == 0 && outl[j].pkttype == 3) {
		if(startflag1 == 0) {   /* Use the first data packet as base  */
		  xmarker = outl[j].y2;
		  strttime = outl[j].y2;
		  startflag1 = 1;
		} else {
		  if(nodrp) {
		    fprintf(fp6,"%f\n",outl[j].y2 - xmarker);
		    mean2 += outl[j].y2 - xmarker;
		    dropcnt++;
		  } else {
		    fprintf(fp5,"%f\n",outl[j].y2 - xmarker);
		    mean1 += outl[j].y2 - xmarker;
		    valcnt1++;
		  }
		  fprintf(fp4,"%f\n",outl[j].y2 - xmarker);
		  mean += outl[j].y2 - xmarker;
		  xmarker = outl[j].y2;
		  valcnt++;
		  if(outl[j].y2 == strttime) {  /* Count back-2-back pkts   */
		    numback++;
		  } else {
		    fprintf(fp7,"%d\n",numback);
		    mean3 += numback;
		    numback = 1;
		    strttime = outl[j].y2;
		    backcnt++;
		  }
		}
	      }
	      if(outl[j].host == 0 && outl[j].pkttype == 6) {
		nodrp = 1;
	      }
	    }
	  startflag = i;
	  lastport = outl[i].rport;
	  startflag1 = 0;
	  nodrp = 0;
	  fprintf(fp7,"%d\n",numback);    /* Don't miss the last back-2-back  */
	  mean3 += numback;
	  backcnt++;
	}
    }
  mean = mean/valcnt;
  mean1 = mean1/valcnt1;
  mean2 = mean2/dropcnt;
  mean3 = mean3/backcnt;
  startflag1 = 0;
  nodrp = 0;
  lastport = outl[0].rport;
  for(i=0;i<num_tln;i++)   /* Now calculate the variances of each data value  */
    {                    
      if(lastport != outl[i].rport || i == num_tln - 1) /* Find file end point*/
	{
	  numback = 1;
	  for(j=startflag;j<i;j++)
	    {
	      if(outl[j].host == 0 && outl[j].pkttype == 3) {
		if(startflag1 == 0) {   /* Use the first data packet as base  */
		  xmarker = outl[j].y2;
		  strttime = outl[j].y2;
		  startflag1 = 1;
		} else {
		  if(nodrp) {
		    varval2 += ((outl[j].y2 - xmarker) - mean2) * 
		      ((outl[j].y2 - xmarker) - mean2);
		  } else {
		    varval1 += ((outl[j].y2 - xmarker) - mean1) * 
		      ((outl[j].y2 - xmarker) - mean1);
		  }
		  varval += ((outl[j].y2 - xmarker) - mean) * 
		    ((outl[j].y2 - xmarker) - mean);
		  xmarker = outl[j].y2;
		  if(outl[j].y2 == strttime) {
		    numback++;
		  } else {
		    varval3 += (numback - mean3) * (numback - mean3);
		    numback = 1;
		    strttime = outl[j].y2;
		  }
		}
	      }
	      if(outl[j].host == 0 && outl[j].pkttype == 6) {
		nodrp = 1;
	      }
	    }
	  startflag = i;
	  lastport = outl[i].rport;
	  startflag1 = 0;
	  nodrp = 0;
	  varval3 += (numback - mean3) * (numback - mean3);
	}
    }
  fprintf(fp3,"Mean of all interpacket spacing:  %f\n",mean);
  fprintf(fp3,"Variance of all interpacket spacing:  %f\n",varval/valcnt);
  fprintf(fp3,"Std. Dev. of all interpacket spacing:  %f\n\n",sqrt(varval/valcnt));
  fprintf(fp3,"Mean of interpacket spacing in slow start:  %f\n",mean1);
  fprintf(fp3,"Variance of interpacket spacing in slow start:  %f\n",varval1/valcnt1);
  fprintf(fp3,"Std. Dev. of interpacket spacing in SS:  %f\n\n",sqrt(varval1/valcnt1));
  fprintf(fp3,"Mean of interpacket spacing in cong. avoid:  %f\n",mean2);
  fprintf(fp3,"Variance of interpacket spacing in cong. avoid:  %f\n",varval2/dropcnt);
  fprintf(fp3,"Std. Dev. of interpacket spacing in CA:  %f\n\n",sqrt(varval2/dropcnt));
  fprintf(fp3,"Mean of back-to-back count lengths:  %f\n",mean3);
  fprintf(fp3,"Variance of back-to-back count lengths:  %f\n",varval3/backcnt);
  fprintf(fp3,"Std. Dev. of back-to-back count lengths:  %f\n\n",sqrt(varval3/backcnt));

  fclose(fp3);
  fclose(fp4);
  fclose(fp5);
  fclose(fp6);
  fclose(fp7);

  /* End function results */
}

/****************************************************************************/
/* This is just a quick comparison routine which is used by qsort.  It      */
/* compares the tdump structure by the rport member.                        */
/****************************************************************************/

int compare1(const void *first, const void *second)
{
  return (((tdump*)first)->rport - ((tdump*)second)->rport);
}

/****************************************************************************/
/* This is just a quick comparison routine which is used by qsort.  It      */
/* compares the tdump structure by the ipid member.  NOTE:  This routine    */
/* should be enhanced with a second index which is  ->host since there can  */
/* be packets with the same ipid sent from each host during a transfer.     */
/****************************************************************************/

int compare2(const void *first, const void *second)
{
  return (((tdump*)first)->ipid - ((tdump*)second)->ipid);
}

/****************************************************************************/
/* This is just a quick comparison routine which is used by qsort.  It      */
/* compares the tline structure by the pkttype member.                      */
/****************************************************************************/

int compare3(const void *first, const void *second)
{
  return (((tline*)first)->pkttype - ((tline*)second)->pkttype);
}

/****************************************************************************/
/* This is just a quick comparison routine which is used by qsort.  It      */
/* compares the tline structure by the y1 (time stamp at client) member.    */
/****************************************************************************/

int compare4(const void *first, const void *second)
{
  return (((tline*)first)->y1long - ((tline*)second)->y1long);
}

/****************************************************************************/
/* This is just a quick comparison routine which is used by qsort.  It      */
/* compares the cpath structure by the dlytype member.                      */
/****************************************************************************/

int compare5(const void *first, const void *second)
{
  return (((cpath*)first)->dlytype - ((cpath*)second)->dlytype);
}

/****************************************************************************/
/* This is just a quick comparison routine which is used by qsort.  It      */
/* compares the tline structure by the host member.                         */
/****************************************************************************/

int compare6(const void *first, const void *second)
{
  return (((tline*)first)->host - ((tline*)second)->host);
}

/****************************************************************************/
/* This is just a quick comparison routine which is used by qsort.  It      */
/* compares the tline structure by the y2 (time stamp at server) member.    */
/****************************************************************************/

int compare7(const void *first, const void *second)
{
  return (((tline*)first)->y2long - ((tline*)second)->y2long);
}

/****************************************************************************/
/* This is just a quick comparison routine which is used by qsort.  It      */
/* compares the tline structure by the host member.                         */
/****************************************************************************/

int compare8(const void *first, const void *second)
{
  return (((tdump*)first)->host - ((tdump*)second)->host);
}

/****************************************************************************/
/* This is just a quick comparison routine which is used by qsort.  It      */
/* compares the cpath structure by the file member.                         */
/****************************************************************************/

int compare9(const void *first, const void *second)
{
  return (((cpath*)first)->file - ((cpath*)second)->file);
}

/****************************************************************************/
/* This is just a quick comparison routine which is used by qsort.  It      */
/* compares the tdump structure keyed first by rport then by time stamps.   */
/****************************************************************************/

int compare10(const void *first, const void *second)
{
  if(((tdump*)first)->rport - ((tdump*)second)->rport > 0)
    {
      return(1);
    } else {
      if(((tdump*)second)->rport - ((tdump*)first)->rport > 0)
        {
          return(-1);
        } else { /* Now key on the ylong values */
	  if(((tdump*)first)->ylong - ((tdump*)second)->ylong > 0)
	    {
	      return(1);
	    } else {
	      if(((tdump*)second)->ylong - ((tdump*)first)->ylong > 0)
		{
		  return(-1);
		} else {
		  return(0);
		}
	    }
        }
    }
}


/****************************************************************************/
/* This functions generates the critical path information for individual    */
/* file transfers.  It populates the cp structure with time contributions   */
/* from either 1) network, 2) server, 3) client, 4) drops or 5) timeout.    */
/* The time contributions are linked together in the form of a time line so */
/* that when printed, they should follow next to the time line for the      */
/* transfer.                                                                */
/*                                                                          */
/* This script uses knowledge of the type of transfer taking place - namely */
/* an HTTP/1.0 transfer (for now).  It assumes that the transfer will       */
/* consist of 1) connection set up, 2) HTTP request, 3) Data transfer and   */
/* 4) connection tear down.                                                 */
/****************************************************************************/

void critpath(long tlstart, long tlend)
{

  int httpstate, founddrop, synstate, getstate, finstate, stopflag, congcnt;
  int l, numdupack, clientmss, droppush, droppop, cli_host, fastretran;
  int droprcvry, cwnd, numdata, lastnumpkts, DEFAULTmss, foundfast;
  int abortflag, oldstopflag, phantomflag;
  float lasty, cp_x, CLNTABORTdelay, SRVRABORTdelay, DROPdelay;
  long i, j, k, sendwin, rndcnt, recvrwin, lastack, datacnt, datacnt1, maxptr;
  long rndnmbrs[MAXSEQ], maxptrrnd, rndindx, k1, dropindx[100], lastdrop;
  long nextack, cp_start, maxrnd, activernd, activepkts, oldactivernd;
  long activecnt, lastdata;
  unsigned long dropseqn[100], finackseq;
  unsigned int ssthresh;

  /* initialize all of the various flags that will be used to track the     */
  /* the progress through the HTTP sequence.                                */

  clientmss = 0;  /* This is set in case 0 assuming it was read in properly */
  httpstate = 0;
  synstate = 0;
  getstate = 0;
  finstate = 0;
  founddrop = 0;
  lasty = 0.0;
  i = 0;
  cli_host = 1;
  cp_x = outl[tlstart].x1 + 1.5;  /* put cp time line between time lines    */

  /* There can be cases when we do not want a transfer to be considered for */
  /* CP analysis.  The two cases are:  1) When the number of packets goes   */
  /* beyond the current recieve window, 2) When the client or server delay  */
  /* observed is beyond the CLNTABORTdelay or SRVRABORTdelay thresholds.    */
  /* Set abortflag = 1 if client delay abort, = 2 if server delay abort     */
  /* = 3 if recieve window abort, = 4 other abort.                          */
  /* We also have a situation where fast retransmit delays are long and     */
  /* must be included on the CP.  If the fast retransmit delay is longer    */
  /* than DROPdelay then include that round on the CP.  If we don't do this */
  /* then some long delays show up as client delays when they are not.      */

  CLNTABORTdelay = 1.75;    /* Client abort delay threshold                 */
  SRVRABORTdelay = 8.0;     /* Server abort delay threshold                 */
  DROPdelay = 0.75;         /* Drop include threshold                       */
  abortflag = 0;            /* Set to 1 when a transfer is aborted          */

  /* Initialize the rounds array and the drops arrays  for the bulk trnsfer */
  for(j=0;j<MAXSEQ;j++)
    {
      xfr[j].phantom = 0;
      xfr[j].ack = 0;
      xfr[j].bytestrt = 0;
      xfr[j].byteend = 0;
      xfr[j].data = 0;
      xfr[j].numpkts = 0;
      xfr[j].founddrop = 0;
      xfr[j].droptype = 0;
      xfr[j].drop = 0;
    }
  for(j=0;j<100;j++) {
    dropindx[j] = 0;
    dropseqn[j] = 0;
  }
  droppush = 0;
  droppop = 0;

  /* begin by sorting the tline array which contains each of the individual */
  /* packet transfers by the time stamp as seen by the client.              */

  qsort(&outl[tlstart],tlend - tlstart,sizeof(outl[0]),compare4);

  /* Now cycle through all of the packets in the transfer.  For each packet */
  /* assign a transfer delay (unless it is dropped) and between packets     */
  /* assign a delay due to either client or server.  Also track the higher  */
  /* http/TCP state as well using the four states listed above.             */

  cp_start = cur_cp;   /* Mark where cur_cp starts in case we must abort    */
  i = tlstart;
  while(i < tlend)
    {
      if(founddrop == 0)
	{
	  cp[cur_cp].x1 = cp_x;            /* set x value for cp time line  */
	  switch(httpstate)
	    {
	    case 0:  /* TCP connection setup sequence - 3 packets           */
	      /* This routine assumes the 3 packets will be in sequential   */
	      /* order in the outl array when sorted by start time.         */

	      /*If pkt is a drop find retransmt but only if it was not an ACK*/
	      if(outl[i].pkttype == 6 && outl[i].origpkt != 2) {
		founddrop = 1;
		dropseqn[0] = outl[i].seqn;
		cp[cur_cp].y1 = outl[i].y1;
		cp[cur_cp].dlytype = 4;  /* Assign delay to type drop       */
		cp[cur_cp].ipid = outl[i].ipid;
		cp[cur_cp].rport = outl[i].rport;
		cp[cur_cp].status = 1;  /* Set status to connection setup   */
		cp[cur_cp].host = outl[i].host;
		cp[cur_cp].file = num_trans;
		if(outl[i].host == 1) {  /* Assign host type to cp          */
		  cp[cur_cp].src = cli_host;
		} 
	      } else {
		if(outl[i].pkttype == 1 && outl[i].host == 1) 
		  {            /* This is the first SYN from the client     */
		    if(synstate == 0) { /*If synstate = 0 its the first SYN */
		      cp[cur_cp].y1 = outl[i].y1;
		      cp[cur_cp].y2 = outl[i].y2;
		      cp[cur_cp].src = cli_host;
		      synstate++;
		    } else {    /* If synstate > 0 its a timeout/retransmit */
		      cp[cur_cp].y1 = lasty;
		      cp[cur_cp].y2 = outl[i].y1;
		      cp[cur_cp].src = cli_host;
		      cp[cur_cp].dlytype = 5; /* Assign delay to retransmit */
		      cp[cur_cp].ipid = outl[i].ipid;
		      cp[cur_cp].rport = outl[i].rport;
		      cp[cur_cp].status = 1;  /* Set status to connetn setup*/
		      cp[cur_cp].host = outl[i].host;
		      cp[cur_cp].file = num_trans;
		      cur_cp++;
		      cp[cur_cp].y1 = outl[i].y1;
		      cp[cur_cp].y2 = outl[i].y2;
		      cp[cur_cp].src = cli_host;
		    }
		    clientmss = outl[i].mss;   /* get client mss for case 2 */
		    cp[cur_cp].x1 = cp_x;  /* set x value for cp time line  */
		    cp[cur_cp].dlytype = 1;  /* assign the network delay    */
		    cp[cur_cp].ipid = outl[i].ipid;
		    cp[cur_cp].rport = outl[i].rport;
		    cp[cur_cp].status = 1;  /*Set status to connection setup*/
		    cp[cur_cp].host = outl[i].host;
		    lasty = cp[cur_cp].y2;
		    cp[cur_cp].file = num_trans;
		    cur_cp++;
		  }
		if(synstate == 1 && outl[i+1].y1 == outl[i+2].y1)/*If SYN-ACK*/
		  { /* arrives at same time as ACK leaves be sure that the  */
		    /* the packet order is SYN-ACK then ACK...              */
		    j = i+1;
		    while(outl[i+1].y1 == outl[j + 1].y1 && j < tlend) j++;
		    qsort(&outl[i+1],j - i,sizeof(outl[0]),compare6);
		    if(j - i == 3)  /* Be sure seq => SYN-ACK, ACK, DATA    */
		      qsort(&outl[i+2],2,sizeof(outl[0]),compare3);
		  }
		if(outl[i].pkttype == 1 && outl[i].host == 0) 
		  {           /*  This is the SYN-ACK from the server       */
		    if(synstate == 1) { /* If synstate == 1 its SYN-ACK time  */
		      if(outl[i].y2 > lasty) {  
			cp[cur_cp].y1 = lasty;
			cp[cur_cp].y2 = outl[i].y2;
			cp[cur_cp].dlytype = 2;   /* Assign delay to server */
			cp[cur_cp].status = 1;  /*Set status to conctn setup*/
			cp[cur_cp].host = outl[i].host;
			cp[cur_cp].rport = outl[i].rport;
			cp[cur_cp].file = num_trans;
			if(outl[i].y2 - lasty > SRVRABORTdelay && abortflag ==0) {
			  printf("CRITPATH:  SRVR Abort transfer case 0 ");
			  printf("Rport %ld IPID %ld Data Seqn %lu\n", 
				 outl[i].rport,outl[i].ipid, outl[i].seqn);
			  abortflag = 2;  /* If delay is too long, abort    */
			}
			cur_cp++;
		      }
		      cp[cur_cp].y1 = outl[i].y2;
		      cp[cur_cp].y2 = outl[i].y1;
		      synstate++;
		    } else {    /* if synstate > 1 its timeout/retransmit   */
		      cp[cur_cp].y1 = lasty;
		      cp[cur_cp].y2 = outl[i].y2;
		      cp[cur_cp].ipid = outl[i].ipid;
		      cp[cur_cp].rport = outl[i].rport;
		      cp[cur_cp].dlytype = 5; /* Assign delay to retransmit */  
		      cp[cur_cp].status = 1;  /* Set status to connctn setup*/
		      cp[cur_cp].host = outl[i].host;
		      cp[cur_cp].file = num_trans;
		      cur_cp++;
		      cp[cur_cp].y1 = outl[i].y2;
		      cp[cur_cp].y2 = outl[i].y1;
		    }
		    cp[cur_cp].x1 = cp_x;  /* set x value for cp time line  */
		    cp[cur_cp].dlytype = 1;  /* Assign the network delay    */
		    cp[cur_cp].status = 1;  /*Set status to connection setup*/
		    cp[cur_cp].host = outl[i].host;
		    cp[cur_cp].ipid = outl[i].ipid;
		    cp[cur_cp].rport = outl[i].rport;
		    lasty = cp[cur_cp].y2;
		    cp[cur_cp].file = num_trans;
		    cur_cp++;
		  }
		if(outl[i].pkttype == 2 && outl[i].host == 1) 
		  {            /*  This is the final ACK from the client    */
		    if(outl[i].y1 > lasty) {
		      cp[cur_cp].y1 = lasty;
		      cp[cur_cp].y2 = outl[i].y1;
		      cp[cur_cp].dlytype = 3;    /* Assign delay to client  */
		      cp[cur_cp].status = 1;  /*Set status to connectn setup*/
		      cp[cur_cp].host = outl[i].host;
		      cp[cur_cp].rport = outl[i].rport;
		      cp[cur_cp].src = cli_host;
		      cp[cur_cp].file = num_trans;
			if(outl[i].y1 - lasty > CLNTABORTdelay && abortflag ==0) {
			  printf("CRITPATH:  CLNT abort transfer case 0 ");
			  printf("Rport %ld IPID %ld Data Seqn %lu\n", 
				 outl[i].rport,outl[i].ipid, outl[i].seqn);
			  abortflag = 1;  /* If delay is too long, abort    */
			}
		      cur_cp++;
		    }
		    cp[cur_cp].x1 = cp_x;  /* set x value for cp time line  */
		    cp[cur_cp].y1 = outl[i].y1;
		    cp[cur_cp].y2 = outl[i].y2;
		    cp[cur_cp].dlytype = 1;  /* Assign the network delay    */
		    cp[cur_cp].status = 1;  /*Set status to connection setup*/
		    cp[cur_cp].host = outl[i].host;
		    cp[cur_cp].ipid = outl[i].ipid;
		    cp[cur_cp].rport = outl[i].rport;
		    cp[cur_cp].src = cli_host;
		    cp[cur_cp].file = num_trans;
		    lasty = cp[cur_cp].y2;
		    lastack = i; /* Set this pointer for the buld xfer case */
		    cur_cp++;
		    httpstate++;  /* Move to HTTP GET state in crit path    */
		  }
	      }
	      break;
	    case 1:  /* This is the HTTP GET requst - usually just one pkt  */
	      /* This routine assumes the GET will be followed either by an */
	      /* ACK or a data packet with a piggybacked ACK                */
	      /* Note that if its just an ACK, then this ACK should not be  */
	      /* on the critical path since data from server is on CP.      */

	      /*If pkt is dropped find retransmt but only if not an ACK     */
	      if(outl[i].pkttype == 6 && outl[i].origpkt != 2 && 
		 outl[i].origpkt != 3) { 
		founddrop = 1;
		dropseqn[0] = outl[i].seqn;
		cp[cur_cp].y1 = outl[i].y1;
		cp[cur_cp].dlytype = 4;  /* Assign delay to type drop       */
		cp[cur_cp].status = 2;  /* Set status to GET/Bulk transfer  */
		cp[cur_cp].host = outl[i].host;
		cp[cur_cp].ipid = outl[i].ipid;
		cp[cur_cp].rport = outl[i].rport;
		cp[cur_cp].file = num_trans;
		if(outl[i].host == 1) {  /* Assign host type to cp          */
		  cp[cur_cp].src = cli_host;
		}
	      } else {
		if(outl[i].pkttype == 3 && outl[i].host == 1) 
		  {             /*  This is the HTTP GET from the client    */
		    if(getstate == 0) { /*If getstate = 0 its the GET reqst */
		      if(outl[i].y1 > lasty) {
			cp[cur_cp].y1 = lasty;
			cp[cur_cp].y2 = outl[i].y1;
			cp[cur_cp].dlytype = 3;  /* Assign delay to client  */
			cp[cur_cp].status = 2;  /*Set status to GET/Bulk xfr*/
			cp[cur_cp].host = outl[i].host;
			cp[cur_cp].rport = outl[i].rport;
			cp[cur_cp].src = cli_host;
			cp[cur_cp].file = num_trans;
			if(outl[i].y1 - lasty > CLNTABORTdelay && abortflag ==0) {
			  printf("CRITPATH:  CLNT abort transfer case 1 ");
			  printf("Rport %ld IPID %ld Data Seqn %lu\n", 
				 outl[i].rport,outl[i].ipid, outl[i].seqn);
			  abortflag = 1;  /* If delay is too long, abort    */
			}
			cur_cp++;
			cp[cur_cp].y1 = outl[i].y1;
			cp[cur_cp].y2 = outl[i].y2;
			cp[cur_cp].x1 = cp_x; /*set x value for cp time line*/
			cp[cur_cp].dlytype = 1;  /*Assign the network delay */
			cp[cur_cp].status = 2;  /*Set status to GET/Bulk xfr*/
			cp[cur_cp].host = outl[i].host;
			cp[cur_cp].ipid = outl[i].ipid;
			cp[cur_cp].rport = outl[i].rport;
			cp[cur_cp].src = cli_host;
			cp[cur_cp].file = num_trans;
			lasty = cp[cur_cp].y2;
			cur_cp++;
		      } else {
			if(outl[i].y2 > cp[cur_cp - 1].y2) { /* add entry   */
			  cur_cp--;                   /*if GET arrived AFTER*/
			  cp[cur_cp].y1 = outl[i].y1; /*last ACK of SYN and */
			  cp[cur_cp].y2 = outl[i].y2; /*replace it on the cp*/
			  cp[cur_cp].dlytype = 1;/*Assign the network delay */
			  cp[cur_cp].status = 2; /*Set status to GET/Bulk xfr*/
			  cp[cur_cp].host = outl[i].host;
			  cp[cur_cp].ipid = outl[i].ipid;
			  cp[cur_cp].rport = outl[i].rport;
			  lasty = cp[cur_cp].y2;
			  cp[cur_cp].src = cli_host;
			  cp[cur_cp].file = num_trans;
			  cur_cp++;
			}
		      }
		      recvrwin = outl[i].win;/*Get recieve window size */
		      getstate++;
		    } else {    /* If getstate > 0 its a timeout/retransmit */
		      cp[cur_cp].y1 = lasty;
		      cp[cur_cp].dlytype = 5; /* Assign delay to retransmit */  
		      cp[cur_cp].status = 2;  /* Set status to GET/Bulk xfr */
		      cp[cur_cp].host = outl[i].host;
		      cp[cur_cp].ipid = outl[i].ipid;
		      cp[cur_cp].rport = outl[i].rport;
		      cp[cur_cp].y2 = outl[i].y1;
		      cp[cur_cp].src = cli_host;
		      cp[cur_cp].file = num_trans;
		      cur_cp++;
		      cp[cur_cp].y1 = outl[i].y1;
		      cp[cur_cp].y2 = outl[i].y2;
		      cp[cur_cp].x1 = cp_x;  /* set x value for cp time line*/
		      cp[cur_cp].dlytype = 1;  /* Assign the network delay  */
		      cp[cur_cp].status = 2;  /* Set status to GET/Bulk xfr */
		      cp[cur_cp].host = outl[i].host;
		      cp[cur_cp].ipid = outl[i].ipid;
		      cp[cur_cp].rport = outl[i].rport;
		      recvrwin = outl[i].win; /* Get recieve window size    */
		      cp[cur_cp].src = cli_host;
		      cp[cur_cp].file = num_trans;
		      lasty = cp[cur_cp].y2;
		      cur_cp++;
		    }
		  }


		/* We can have a situation with small files where data, fin */
		/* fin-ack and fin all have the same time stamp at client   */
		/* so we must be sure that the pkt order is correct         */
		if(getstate == 1 && outl[i+1].y1 == outl[i+2].y1 &&
		   outl[i+1].y1 == outl[i+3].y1) /* need to check next 3 pkt*/
		  { 
		    /* be sure order id {HTTP ACK}, DATA, FIN, ACK, FIN     */
		    /* so first sort by host and then by pkt type           */
		    j = i + 1;
		    while(outl[i + 1].y1 == outl[j + 1].y1 && j < tlend) j++;
		    qsort(&outl[i+1],j - i,sizeof(outl[0]),compare6);
		    j = i + 1;
		    while((outl[i + 1].y1 == outl[j + 1].y1) &&
			  (outl[i + 1].host == outl[j + 1].host) &&
			  j < tlend) j++;
		    if(j - i >= 2)  /* Be sure seq => DATA FIN from server  */
		      qsort(&outl[i + 1],j - i,sizeof(outl[0]),compare3);
		    l = j + 1;
		    while((outl[i + 1].y1 == outl[l + 1].y1) &&
			  (outl[j + 1].host == outl[l + 1].host) && 
			  j < tlend) j++;
		    if(l - j >= 2)  /* Be sure seq => FIN ACK from client   */
		      qsort(&outl[j + 1],l - j,sizeof(outl[0]),compare3);
		    getstate++;
		  }
		if((outl[i].pkttype == 2 || outl[i].pkttype == 3) 
		   && outl[i].host == 0) 
		  { /* This is ACK of HTTP GET from the server or first data*/
		    if(outl[i].pkttype == 2) { /* ACK is not piggybacked    */
		      /* Actually, This ACK SHOULD NOT be on CP ever... PB  */
		      /**********************************************
		        if(outl[i].y2 > lasty) {
			  cp[cur_cp].y1 = lasty;
			  cp[cur_cp].y2 = outl[i].y2;
			  cp[cur_cp].dlytype = 2;
			  cp[cur_cp].status = 2;
			  cp[cur_cp].host = outl[i].host;
			  cp[cur_cp].rport = outl[i].rport;
			  cp[cur_cp].file = num_trans;
			  if(outl[i].y2 - lasty > SRVRABORTdelay && abortflag ==0) {
			    printf("CRITPATH:  SRVR abort transaction case 1 ");
			    printf("Rport %ld IPID %ld Data Seqn %lu\n", 
				 outl[i].rport,outl[i].ipid, outl[i].seqn);
			    abortflag = 2;
			  }
			  cur_cp++;
		        }
		        cp[cur_cp].x1 = cp_x;
		        cp[cur_cp].y1 = outl[i].y2;
		        cp[cur_cp].y2 = outl[i].y1;
		        cp[cur_cp].dlytype = 1;
		        cp[cur_cp].status = 2;
		        cp[cur_cp].host = outl[i].host;
		        cp[cur_cp].ipid = outl[i].ipid;
		        cp[cur_cp].rport = outl[i].rport;
		        lasty = cp[cur_cp].y2;
		        cp[cur_cp].file = num_trans;
		        cur_cp++;
		      *************************************************/
		    } else { /* Piggybacked ACK so consider this pkt below  */
		      i--;
		    }
		    httpstate++; /*Move to transfer data state in crit path */
		  }
		/* If piggybacked ACK was dropped, consider this pkt below  */
		if(outl[i].pkttype == 6 && outl[i].origpkt == 3) {
		    if(outl[i].y2 > lasty) {
		      cp[cur_cp].y1 = lasty;
		      cp[cur_cp].y2 = outl[i].y2;
		      cp[cur_cp].dlytype = 2;  /*  Assign delay to server   */
		      cp[cur_cp].status = 2;  /* Set status to GET/Bulk xfr */
		      cp[cur_cp].host = outl[i].host;
		      cp[cur_cp].rport = outl[i].rport;
		      cp[cur_cp].file = num_trans;
		      if(outl[i].y2 - lasty > SRVRABORTdelay && abortflag == 0) {
			printf("CRITPATH:  SRVR abort transaction case 1 ");
			printf("Rport %ld IPID %ld Data Seqn %lu\n", 
			       outl[i].rport,outl[i].ipid, outl[i].seqn);
			abortflag = 2;  /* If delay is too long, abort    */
		      }
		      cur_cp++;
		    }
		    i--;
		    httpstate++; /*Move to transfer data state in crit path */
		}
	      }
	      break;

	      /* Now case 2 is where all of the fun starts.  This is the    */
	      /* response to the GET request and is where most of the data  */
	      /* transfer takes place and where critical path is hardest to */
	      /* model.  We model critical path by first dividing the pkt   */
	      /* transfer sequence into "rounds" where each round is some   */
	      /* number of data packets from the server followed by an ACK  */
	      /* from the client. This enables us to track a forward        */
	      /* dependance relation between data and ACKS.  We then go     */
	      /* backward through the rounds making a backward link between */
	      /* ACK and data packet based on seq. no.  From here we can    */
	      /* construct the critical path from initial data to final data*/

	    case 2:

	      /* Now sort the bulk transfer part of the transaction by time */
	      /* stamp at the server since this is where we best track cp   */

	      j = tlend;
	      qsort(&outl[i],j - i,sizeof(outl[0]),compare7);

	      if(DEBUG==4) for(k=i;k<j;k++) 
		printf("Data frame:  pkttype = %d seq = %lu Time = %ld ipid = %ld\n",
		       outl[k].pkttype,outl[k].seqn, outl[k].y2long, outl[k].ipid);

	      /* Do first pass through bulk transfer pkts and set up rounds */
	      /* Send window must be tracked per TCP in order to set rounds */
	      /* up correctly.  Round = data followed by ACK where each ACK */
	      /* triggers as much data as the TCP algorithm allows. Clearly */
	      /* this means that this part of the code must be set up to    */
	      /* model the specific TCP implementation that is being used   */
	      /* at the point where the measurements are being taken.       */
              /* The tricky part here is dealing with byte start and byte   */
              /* Byte start will always be the next byte after last byte end*/
              /* Byte end will be the number of bytes transmitted in the    */
              /* next n data packets where n is the number of packets which */
              /* can be sent per TCP.                                       */

	      rndcnt = 0;              /* # rounds in this data transfer    */
	      xfr[rndcnt].ack = i - 1;  /* Last ACK triggers bulk transfer  */
	      xfr[rndcnt].data = i - 1; /* Just a dummy data index...       */
	      xfr[rndcnt].bytestrt = 1;/* byte count start for first round  */
	      lastack = 1;   /* The last ack from the client is the synack  */
	      numdupack = 0;           /* # dup ACK pkts seen               */
	      congcnt = 0;             /* # of pkts seen in cong avoid mode */
	      cwnd = 2;      /* 2 since we start with data part of HTTP seq */

	      /* NOTE AS FAR AS TCPEVAL IS CONCERNED, the following parms   */
	      /* are the only ones which should be unique for a given TCP   */
	      /* implementation.  This assumes TCP uses standard SYN, bulk  */
	      /* transfer, FIN sequence where bulk transfer assumes sliding */
	      /* window, slow start, fast retran, fast recovery only.       */

	      ssthresh = 0x7fffffff; /* Linux 2.0.30 starts with this value */
	      DEFAULTmss = 536;      /* TCP default mss size                */

	      if(clientmss == 0)     /* clientmss should be set in case 0   */
		{
		  printf("CRITPATH:  Client mss = 0, Setting mss to default\n");
		  clientmss = DEFAULTmss;
		}

	      /* Set send win = min of advertised window and cong window    */
	      if(recvrwin < cwnd * clientmss) {
		xfr[rndcnt].numpkts = recvrwin/clientmss;
		lastnumpkts = recvrwin/clientmss;
	      } else {
		xfr[rndcnt].numpkts = cwnd;  /* # packets released by ACK   */
		lastnumpkts = cwnd;  /* # packets which COULD be released   */
	      }

	      /* Set up byte boundaries for data triggered by ACK's         */
	      /* This code is specific to the TCP implementation used...    */
	      /* We have to count bytes sent in data packets in order to    */
	      /* track the recieve window properly.                         */
	      /* This loop also assigns a type to each dropped packet that  */
	      /* appears in a round (either time out (5) or fast retransmit */
	      /* (4).   NOTE that whenever there is a drop, the number of   */
	      /* pkts for the past round must be changed unless its the last*/
	      /* pkt which is lost - change to the actual number sent.  The */
	      /* number of pkts for the round after the drop is sent to the */
	      /* actual number which were sent - this may not be right...   */

	      phantomflag = 0;  /* Flag indicating phantom drop found       */
	      droprcvry = 0;    /* Flag indicating time out drop data found */
	      fastretran = 0;   /* Flag indicating if the drop is a fast rtn*/
	      foundfast = 0;    /* Flag indicating fast rtrn data pkt found */
	      sendwin = 0;           /* Number of bytes of data in transit  */
	      activernd = 0;  /* track which rnd is currently genratng data */
	      activepkts = xfr[activernd].numpkts; /* # pkts for current rnd*/
	      activecnt = 0;  /* Count of active pkts sent in active rnd    */
	      numdata = 0;    /* Number of packets sent between current rnds*/
	      maxptr = i; /* points to pkt with largest seqn in all rounds  */
	      for(k=i;k<j;k++)  /* Read through the rest of the entire xfer */
		{
		  /* Track ACK's to set up new rounds.  Only consider ACKs  */
		  /* which have seqn greater than or equal to lastack...    */
		  if(outl[k].pkttype == 2 && outl[k].host == 1 && 
		     outl[k].seqn >= lastack) {
		    /* NOTE:  Cwnd and ssthresh are maintained in pkts NOT  */
		    /* in bytes - to be like TCP spec and Stevens book, both*/
		    /* should be maintained in bytes...but my code works :-)*/
		    if(cwnd <= ssthresh) /* This is standard TCP cong win   */
		      {	        /* slow start so increment cwnd on each ACK */
			cwnd++;
		      } else {  /* congest avoid so increment 1/cwnd on ACK */ 
			if(congcnt >= cwnd)
			  {
			    cwnd++;
			    congcnt = 0;
			  } else {
			    congcnt++;
			  }
		      }
		    if(outl[k].seqn == lastack)  /* Is this a duplicate ACK?*/
		      {
			numdupack++;       /* No new round on duplicate ACK */
			if(DEBUG == 4) 
			  printf("Duplicate ACK # %d found:  Seq # = %lu\n",
				 numdupack,outl[k].seqn);
			/* If there have been dup 3 ACK's then we're in     */
			/* fast retransmit so set ssthresh to cwnd/2 or 2   */
			if(numdupack == 3 && 
			   outl[dropindx[droppop]].droptype == 0) { 
			  if(cwnd/2 < 2) {   /* reset ssthresh to max...    */
			    ssthresh = 2;    /* per fast retrans, fast rcvry*/
			  } else {
			    ssthresh = cwnd/2;
			  }
			  cwnd = ssthresh + 3;      /* Per standard TCP     */
			  /* Set drop type to fast retransmit on 3rd dup ack*/
			  /* but only flag as such if we have seen this drop*/
			  if(droppush != droppop) {
			    outl[dropindx[droppop]].droptype = 4;
			  } else {   /*Drop signaled but not seen ie. phantom*/
			    foundfast = 1; 
			    phantomflag = 1; /* Set fst rtn phantom flag    */
			  }
			  fastretran = 1;  /* Set flag to do fast recovery  */
			  if(DEBUG==4) 
			    printf("Fast Retrans Signal:  Seq # = %lu\n",
					dropseqn[droppop]);
			  if(xfr[activernd].numpkts == 0) /* Enable lost    */
			    xfr[activernd].numpkts++;   /* pkt to be resent */
			}
			if(numdupack > 3) { /* Additional dup ack after fst-*/
			  cwnd++;   /* retrans causes cwnd to increase by 1 */
			}
		      } else { /* If not a dup ACK then start a new round */
			if(fastretran && foundfast) {  /* Do fast recovery  */
			  if(DEBUG==4) 
			    printf("Fast Recovery on ACK of data  Seq # = %lu\n",
				   outl[k].seqn);
			  cwnd = ssthresh; /* So we're in cong. avoidance*/
			  foundfast = 0;
			  fastretran = 0;
			}
			/* We can have a case of a phantom time out drop.   */
			/* This is when there is a transmit of a data with  */
			/* seqn < ACK seqn and no drop was detected.  If    */
			/* found then adjust windows and set flags...       */
			if(outl[lastdata].seqn + 1 < outl[k].seqn && 
			   !droprcvry && phantomflag == 0) {
			  phantomflag = 2;
			  xfr[rndcnt+1].phantom = 2;
			  /* Adjust  windows and active rounds just like a  */
			  /* normal time out drop...                        */
			  if(cwnd/2 < 2) {   /* Set ssthresh per TCP        */
			    ssthresh = 2;
			  } else {
			    ssthresh = cwnd/2;
			  }
			  cwnd = 1;                   /* Set cwnd per TCP   */
			  oldactivernd = activernd;   /* Adjust rounds      */
			  xfr[activernd].numpkts = activecnt;
			  numdata = 0;
			  if(rndcnt > activernd) {
			    activernd++;
			    while(activernd < rndcnt) {
			      xfr[activernd].numpkts = 0;
			      activernd++;
			    }
			  }
			  if(DEBUG==4) 
			    printf("Found Phantom Time Out:  Assigned to ACK # = %lu\n",
					outl[k].seqn);
			}
			/* If last round had a time out drop then adjust    */
			/* numpkts to the actual number of pkts transfered  */
			if((droprcvry && 
			   outl[dropindx[droppop - 1]].droptype == 5) ||
			   phantomflag == 2) {
			  /* The next if statement adjusts the rnd in which */
			  /* the drop occured to the actual number of pkts  */
			  if(oldactivernd < rndcnt) {
			    xfr[rndcnt].numpkts = numdata;
			  } else {
			    xfr[rndcnt].numpkts += numdata;
			  }
			  activecnt = 0;
			  numdata = 0;
			  sendwin = 0;
			  lastnumpkts = ceil(((float)outl[maxptr].seqn - 
					      (float)outl[k].seqn)/(float)clientmss);
			}
			rndcnt++;                /* Increment round counter */
			numdupack = 0;           /* Reset dup ack counter   */
			if(sendwin > 0)    /* Adjust sendwin for ACKed data */
			  sendwin -= outl[k].seqn - lastack;
			if(sendwin < 0) sendwin = 0; /* sndwin can't be < 0 */
			lastack = outl[k].seqn;         /* Set new lastack  */
			xfr[rndcnt].ack = k;     /* Assign ACK index to xfr */
			/* Now set the number of packets which are released */
			/* by this ACK which depends on whether we are cwnd */
			/* or recvr win limited.  Set to window size minus  */
			/* the number of packets currently in flight        */
			if(recvrwin < cwnd * clientmss ) {
			    xfr[rndcnt].numpkts = 
			      ((recvrwin - sendwin)/clientmss) - 
			      (lastnumpkts - numdata);
			    lastnumpkts = (recvrwin - sendwin)/clientmss;
			} else {
			  xfr[rndcnt].numpkts = (cwnd - 
			    ceil((float)sendwin/(float)clientmss)) -
			    (lastnumpkts - numdata);
			  lastnumpkts = cwnd - 
			    ceil((float)sendwin/(float)clientmss);
			}
			/* Just an extra check to be sure numpkts for rnd  */
			/* is within limits of 0 and reciever window       */
			if(xfr[rndcnt].numpkts < 0) xfr[rndcnt].numpkts = 0;
			if(xfr[rndcnt].numpkts > recvrwin/clientmss)
			  xfr[rndcnt].numpkts = recvrwin/clientmss;
			if(recvrwin > cwnd * clientmss && 
			   xfr[rndcnt].numpkts > cwnd)
			  xfr[rndcnt].numpkts = cwnd;

			/* If there was a phantom drop (one not seen in     */
			/* traces but signaled via 3-dup ack, then set flag */
			if(phantomflag == 1) {
			  xfr[rndcnt].phantom = 1;
			  phantomflag = 0;
			  if(DEBUG==4) 
			    printf("Found Phantom Drop:  Assigned to ACK # = %lu\n",
					outl[xfr[rndcnt].ack].seqn);
			}

			/* Finish adjustments to active round counters    */
			/* after a timeout drop...                        */
			if((droprcvry && 
			   outl[dropindx[droppop - 1]].droptype == 5) ||
			   phantomflag == 2) {
			  activernd = rndcnt;
			  activepkts = xfr[rndcnt].numpkts;
			  droprcvry = 0;
			  phantomflag = 0;
			}

			/* The active round can only be at most the max    */
			/* window size behind the current round in terms of*/
			/* the number of pkts which can be transmitted     */
			stopflag = 0;
			l = rndcnt;
			while(stopflag < recvrwin/clientmss && 
			      stopflag < cwnd && l >= activernd) {
			  stopflag += xfr[l].numpkts;
			  l--;
			}
			if(stopflag > recvrwin/clientmss || stopflag > cwnd) 
			  l++;
			while(activernd < l) { /* If active rnd is advanced */
			  xfr[activernd].numpkts = 0;  /* zero out unused   */
			  activernd++;                 /* rounds            */
			}

			/* If we think there is a drop and the ACK seqn is  */
			/* greater than the seqn of the packet we thought   */
			/* was dropped, then we conclude that the packet    */
			/* filter dropped the packet and adjust drop data.  */
			/* There is also a weird case when a retran is      */
			/* dropped even though a prior retran was received. */
			if(founddrop > 0)
			  while(founddrop > 0 && 
				outl[k].seqn > dropseqn[droppop]) {
			    founddrop--;
			    droppop++;
			  }
			numdata = 0;
			if(DEBUG == 4) {
			  printf("RND: %ld ACK: %lu CWND: %d",rndcnt,lastack,cwnd);
			  printf(" SENDW: %f",ceil((float)sendwin/(float)clientmss));
			  printf(" NUMPKTS: %d",xfr[rndcnt].numpkts);
			  printf(" ACTIVERND: %ld\n",activernd);
			}
		      }
		  }

		  /* Track first FIN so that we don't inject a phantom drop */

		  if(outl[k].pkttype == 4 && outl[k].host == 0) {
		    lastdata = k;
		  }

		  /* Track data in order to track the receiver window size  */

		  if(outl[k].pkttype == 3 && outl[k].host == 0 && 
		     outl[k].seqn >= lastack) {

		    numdata++;        /* Increment pkt count between rounds */
		    activecnt++;      /* Increment num active pkt cnt...    */
		    /* Find the largest data packet which was transferred   */
		    if(outl[k].seqn > outl[maxptr].seqn) maxptr = k;

		    /* Check to see if we have found a retransmitted packet */
                    if(founddrop > 0 && dropseqn[droppop] == outl[k].seqn) {
		      /* If a retransmit was caused by a time out, set cwnd */
		      /* to 1 and restart slow start                        */
		      if(outl[dropindx[droppop]].droptype != 4) {
			if(cwnd/2 < 2) {   /* Set ssthresh per TCP          */
			  ssthresh = 2;
			} else {
			  ssthresh = cwnd/2;
			}
			cwnd = 1;                      /* Set cwnd per TCP  */
			outl[dropindx[droppop]].droptype = 5; /* time out   */
			if(DEBUG==4) 
			  printf("Time out drop retrans found:  Seq # = %lu\n",
					      dropseqn[droppop]);
			/* Adjust the number of pkts sent in active round   */
			/* and zero out the pkts triggered by ACK's         */
			/* between active round and rndcnt.  The move       */
			/* active round pointer up to rndcnt.               */
			oldactivernd = activernd;
			xfr[activernd].numpkts = activecnt;
			numdata = 0;
			if(rndcnt > activernd) {
			  activernd++;
			  while(activernd < rndcnt) {
			    xfr[activernd].numpkts = 0;
			    activernd++;
			  }
			}
			numdupack = 3; /* Prevent a TO being called a retran*/
		      } else { /* If its not a TMO then its a fast retrans  */
			if(DEBUG==4) 
			  printf("Fast retrans found:  Seq # = %lu\n",
					      dropseqn[droppop]);
			foundfast = 1;
		      }
		      droprcvry = 1;    /* Set flag to adjust num pkts sent */
		      founddrop--;;      /* dcrmnt found drop flag for loop */
		      droppop++;         /* incrmnt pop pntr to nxt drop    */
		    } else {
		      stopflag = 1;
		      for(l=0;l<=droppop;l++)
			if(outl[k].seqn == dropseqn[l]) stopflag = 0;
		      if(stopflag) { /*Only add to sendwin if its not a resend*/
			sendwin += outl[k].bytes;  /* Add data sent to sendwin*/
			if(recvrwin + clientmss < sendwin && abortflag == 0) {
			  printf("CRITPATH:  Data bytes sent beyond recv win ");
			  printf("Rport %ld IPID %ld Data Seqn %lu RCVW %ld SNDW %ld\n", 
				 outl[k].rport,outl[k].ipid, outl[k].seqn,recvrwin,
				 sendwin);
			  abortflag = 3;
			}
		      }
		    }
		    /* Adjust active pkt count for data even if its a retrns*/
		    /* Active pkts are initialized to cwnd and track the    */
		    /* data triggered by each ACK so that when a drop is    */
		    /* found, in-flight data can be associated with an ACK  */
		    if(activepkts == 0 && activernd == rndcnt && 
		       droprcvry == 0) {
		      xfr[rndcnt].numpkts++;
		    }
		    if(activepkts == 0 && activernd < rndcnt) {
		      activernd++;
		      while(xfr[activernd].numpkts == 0 && activernd <= rndcnt)
			activernd++;
		      activepkts = xfr[activernd].numpkts;
		      if(activepkts > 0) activepkts -= 1;
		      activecnt = 1;
		    } else {
		      if(activepkts > 0) activepkts--;
		    }

		    lastdata = k;     /* Set last data pointer to current   */
		    if(DEBUG==4)
		      printf("Data seqn = %lu activernd = %ld activepkts = %ld\n",
			     outl[k].seqn, activernd, activepkts);
		  }
		  /* Track drops to tell if if drop is indicated by time out*/
		  /* There can be multiple drops so track them all!         */

		  if(outl[k].pkttype == 6 && outl[k].origpkt == 3  
		     && outl[k].host == 0) {

		    numdata++;
		    activecnt++;      /* Increment num active pkt cnt...    */
		    /* If this drop is the drop of a resend then don't make */
		    /* another entry for it in the drop array.              */
		    stopflag = 1;
		    if(outl[k].seqn == dropseqn[droppop]) stopflag = 0;
		    if(stopflag) {
		      founddrop++;             /* Set drop packet index and */
		      dropindx[droppush] = k;  /* flags for the round       */
		      dropseqn[droppush] = outl[k].seqn;
		      droppush++;
		      if(droppush == 100 && abortflag == 0) {
		      printf("TOO MANY DROPS IN THIS FILE TRANSFER, Port %ld\n",
			     outl[k].rport);
			abortflag = 4;
		      }
		      /* Adjust the recv window as normal since from the    */
		      /* senders perspective, this is a normal packet       */
		      sendwin += outl[k].bytes; /* Increment recv window    */
		      if(recvrwin + clientmss< sendwin && abortflag == 0)  {
			printf("CRITPATH:  Dropped Bytes sent beyond recv win ");
			printf("Rport %ld IPID %ld Data Seqn %lu\n", 
			       outl[k].rport,outl[k].ipid,outl[k].seqn);
			abortflag = 3;
		      }
		    } else {
		      if(DEBUG==4)
			printf("Found drop of a resend: Seqn # %lu IPID %ld\n",
			       outl[k].ipid,outl[k].seqn);
		      /*If we have dropped a fast retransmitted packet then */
		      /* change its type to a time out since that's how it  */
		      /* will be detected next.                             */
		      outl[dropindx[droppop]].droptype = 5;
		      fastretran = 0;
		    }

		    /* Adjust active pkt count even if the pkt is dropped   */

		    if(activepkts == 0 && activernd == rndcnt) {
		      xfr[rndcnt].numpkts++;
		    }
		    if(activepkts == 0 && activernd < rndcnt) {
		      activernd++;
		      while(xfr[activernd].numpkts == 0 && activernd <= rndcnt)
			activernd++;
		      activepkts = xfr[activernd].numpkts;
		      if(activepkts > 0) activepkts -= 1;
		      activecnt = 1;
		    } else {
		      if(activepkts > 0) activepkts--;
		    }

		    if(DEBUG==4)
		      printf("Drop seqn = %lu activernd = %ld activepkts = %ld\n",
			     outl[k].seqn, activernd, activepkts);
		  }
		}  /* End of bulk transfer round set up loop                */

	      if(DEBUG==4)
		for(k=0;k<=rndcnt;k++)
		  printf("Round %ld, ACK = %lu, NumPkts = %d\n", 
			 k, outl[xfr[k].ack].seqn,xfr[k].numpkts);

	      /* Now go through data again and find the last data packet    */
	      /* sent in a round so that we can set up the bytestrt and     */
	      /* byteend in this loop by counting numpkts of data packets   */
	      /* for each round and assigning the number of bytes           */
	      /* transferred to byteend.  This loop is separate from the    */
	      /* loop above since our TCP above deals in packets not bytes  */
	      /* however it would be better to just omit pkts all together  */
	      /* NOTE:  Be careful in this loop since bytestart cannot be   */
	      /* less than the ACK seqn and byteend must always be greater  */
	      /* than bytestart.                                            */

	      maxptr = i; /* points to pkt with largest seqn in all rounds  */
	      maxptrrnd = i; /* points to pkt with largest seqn in a round  */
	      rndindx = 0;
	      stopflag = 0;
	      oldstopflag = 0;
	      for(k=i;k<j;k++)
		if((outl[k].pkttype == 3 || outl[k].origpkt == 3)  
		   && outl[k].host == 0) {
		  /* Find the largest data packet which was transferred     */
		  if(outl[k].seqn > outl[maxptr].seqn) maxptr = k;
		  /* Find largest seqn for the round to set byteend         */
		  if(outl[k].seqn > outl[maxptrrnd].seqn) maxptrrnd = k;
		  xfr[rndindx].numpkts--;
		  if(xfr[rndindx].numpkts == 0) {
		    if(rndindx != 0)   /* Set bytestart after the first rnd */
		      xfr[rndindx].bytestrt = xfr[stopflag].byteend;
		    xfr[rndindx].byteend = outl[maxptrrnd].seqn; /* Set end */
		    maxptrrnd = i;
		    stopflag = rndindx;
		    /* Find the next round with a non-zero number data pkts */
		    while(xfr[rndindx].numpkts == 0 && rndindx < rndcnt) 
		      rndindx++;
		  }
		}   /* End of data packet assignment loop                   */

	      /* Now we have to adjust the byte boundries since due to      */
	      /* retransmits, ends and starts can get out of synch...       */

	      k = 1;
	      maxrnd = 0;
	      while(k < rndindx) {
		if(xfr[k].bytestrt != 0) {
		  if(xfr[k].bytestrt < outl[xfr[k].ack].seqn)
		    xfr[k].bytestrt = outl[xfr[k].ack].seqn;
		  if(xfr[k].bytestrt != xfr[maxrnd].byteend)
		    xfr[maxrnd].byteend = xfr[k].bytestrt;
		  if(xfr[k].byteend <= xfr[k].bytestrt && k < rndindx - 1)
		    xfr[k].byteend = outl[xfr[k + 1].ack].seqn;
		  if(xfr[k].byteend <= xfr[k].bytestrt) {
		    xfr[k].bytestrt = 0;
		    xfr[k].byteend = 0;
		  } else {
		    maxrnd = k;
		  }
		}
		k++;
	      }     /* End of data packet adjustment loop                   */

	      if(DEBUG==4) 
		for(k=0;k<rndindx;k++)
		  printf("Round %ld, ACK = %lu, Bytestart = %ld Byteend = %ld\n", 
			 k, outl[xfr[k].ack].seqn, xfr[k].bytestrt, 
			 xfr[k].byteend);


	      /* Set the byte start and byte end for the last round         */
	      /* Assume that the last pkt in sequence is an ACK from        */
	      /* client of FIN from server                                  */
	      if(rndindx == rndcnt && numdupack == 0) rndindx--;
	      if(outl[xfr[rndindx].ack].seqn > xfr[rndindx-1].byteend &&
		 rndindx != 0) {
		xfr[rndindx].bytestrt = outl[xfr[rndindx].ack].seqn;
		xfr[stopflag].byteend = outl[xfr[rndindx].ack].seqn;
	      } else {
		xfr[rndindx].bytestrt = xfr[rndindx-1].byteend;
	      }
	      xfr[rndindx].byteend = outl[maxptr].seqn;
	      if(DEBUG==4) 
		printf("FRnd %ld, ACK = %lu, Bytestart = %ld Byteend = %ld\n",
		       rndindx, outl[xfr[rndindx].ack].seqn, 
		       xfr[rndindx].bytestrt, xfr[rndindx].byteend);

	      /* Now do a quick loop through the rounds to trim down the    */
	      /* number of rounds.  This is done because after we find the  */
	      /* round in which the last data byte is transfered, all of the*/
	      /* remaining rounds will not be on the critical path.         */

	      maxrnd = rndcnt;
	      datacnt = 0;
	      stopflag = 1;
	      while(stopflag)
		if (xfr[datacnt].byteend < outl[maxptr].seqn && 
		    datacnt <= rndindx) {
		  datacnt++;
		} else {
		  stopflag = 0;
		}
	      rndindx = datacnt;
	      if(DEBUG==4) printf("Round index = %ld\n",rndindx);

	      /* The result of this loop is to find the data which triggered*/
	      /* the ACK:  it can be the data with the same seq num as the  */
	      /* ACK or an out of order data pkt or a retransmitted pkt.    */
	      /* This loop also sets up the found drop flag in a round.     */
	      /* There can only be one drop in a round and this flag is     */
	      /* set when the retransmitted packet for a drop is found      */

	      rndcnt = 1;
	      founddrop = 0;
	      droppush = 0;
	      for(l=0;l<100;l++) dropseqn[l] = 0;
	      for(k=i;k<j;k++) {
		/* Make data assignment for retransmitted packets which     */
		/* trigger an ACK. Also set drop flag for round where the   */
		/* retransmitted packet is found.                           */
		if(outl[k].pkttype == 6 && outl[k].origpkt == 3 
		   && outl[k].host == 0) {

		  /* Once again we have to account for multiple drops of    */
		  /* the same pkt - ignore if we've already seen this drop  */
		  stopflag = 1;
		  for(l=0;l<droppush;l++)
		    if(outl[k].seqn == dropseqn[l]) stopflag = 0;
		  if(stopflag) { /* First time seeing this drop             */
		    dropseqn[droppush] = outl[k].seqn;
		    droppush++;
		    if(droppush == 100 && abortflag == 0) {
		      printf("TOO MANY DROPS IN THIS FILE TRANSFER, Port %ld\n",
			     outl[k].rport);
		      abortflag = 4;
		    }
		    lastdrop = k;
		    founddrop = 1;
		    if(DEBUG==4) 
		      printf("Found drop with seq # %lu IPID = %ld\n",
			outl[k].seqn,outl[k].ipid);
		    dropindx[0] = k;
		    /* Now find the retran of the drop that we just found   */
		    datacnt = k + 1;
		    while(founddrop && datacnt < j) {
		      if(outl[datacnt].pkttype == 3 && 
			 outl[dropindx[0]].seqn == outl[datacnt].seqn) {
			lastdrop = datacnt;
		      if(DEBUG==4) 
			printf("Found retrans of drop with seq # %lu IPID %ld\n",
			       outl[datacnt].seqn, outl[datacnt].ipid);
		      founddrop = 0;
		      }
		      datacnt++;
		    }
		    /* When we find the retransmit, the next ACK is the ACK */
		    /* for the retrnsmted packet. Skip if retrans not found */
		    if(founddrop == 0) {
		      founddrop = 1;
		      while(founddrop && datacnt < j) {
			if(outl[datacnt].pkttype == 2 && 
			   outl[datacnt].seqn >= outl[lastdrop].seqn) {
			  nextack = datacnt;
			  if(DEBUG==4) 
			    printf("Found ACK of retrans with  seq # %lu IPID %ld\n",
				   outl[datacnt].seqn, outl[datacnt].ipid);
			  founddrop = 0;
			}
			datacnt++;
		      }
		    }
		    if(founddrop) {
		      printf("No Retrans or match for ACK of drop seq # %lu port %ld\n",
			     outl[lastdrop].seqn,outl[lastdrop].rport);
		      founddrop = 0;
		    } else { /* Now assign data and drop to the right round */
		      datacnt = 0;
		      founddrop = 1;
		      while(founddrop && datacnt <= maxrnd) {
			if(outl[xfr[datacnt].ack].seqn == outl[nextack].seqn) {
			  xfr[datacnt].data = lastdrop;
			  xfr[datacnt].founddrop = 1;
			  xfr[datacnt].drop = dropindx[0];
			  xfr[datacnt].droptype = outl[dropindx[0]].droptype;
			  if(xfr[datacnt].droptype == 0) xfr[datacnt].droptype = 5;
			  if(DEBUG==4) 
			    printf("Round %ld contains ACK seq # %lu DropTyp = %d\n",
				   datacnt, outl[nextack].seqn,xfr[datacnt].droptype);
			  while(xfr[rndcnt].data != 0 && rndcnt < maxrnd) rndcnt++;
			  founddrop = 0;
			}
			datacnt++;
		      }
		      if(founddrop) {
			printf("No round match for ACK of drop seq: # %lu port %ld\n",
			       outl[nextack].seqn,outl[nextack].rport);
			if(abortflag == 0) abortflag = 4;
			founddrop = 0;
		      }
		    }
		  }
		}  /* End if for test if the packet was a drop              */

		/* Now make data assignment for normal packets.  Be sure not*/
		/* to overwrite a retransmit data already assigned to a rnd */
		if(rndcnt <= maxrnd && outl[k].pkttype == 3 && 
		   outl[k].seqn == outl[xfr[rndcnt].ack].seqn &&
		   xfr[rndcnt].data == 0  && outl[k].host == 0) {
		  xfr[rndcnt].data = k;
		  while(xfr[rndcnt].data != 0 && rndcnt < maxrnd) rndcnt++;
		} 
		/* Now deal with the case of badly out of order data packets */
		/* Assign data to last data received before this rndcnt ack  */
		if(outl[k].pkttype == 2 && 
		   outl[k].seqn > outl[xfr[rndcnt].ack].seqn) {
		  k = xfr[rndcnt].ack; /* Reset k to proper position to proceed */
		  while(outl[k].pkttype != 3 && k > i) k--;
		  xfr[rndcnt].data = k;
		  k = xfr[rndcnt].ack; /* Reset k to proper position to proceed */
		  while(xfr[rndcnt].data != 0 && rndcnt < maxrnd) rndcnt++;
		} 
	      }  /* End of drop check loop                                  */

	      /* Now test for and assign phantom drops.  There are cases    */
	      /* where there is a drop signaled by 3-dup ACK's for pkts     */
	      /* which do not show up as drops in the traces.  Assign       */
	      /* these as drops in the critical path.  There can also be    */
	      /* phantom time out drops assign these in the same way.       */

		for(k=0;k<=rndindx;k++) {
		  if(xfr[k].phantom > 0 && xfr[k].founddrop == 0) {
		    stopflag = 1;
		    datacnt1 = xfr[k].ack - 1; /*Find data which triggerd ACK*/
		    while(stopflag && datacnt1 >= i) {
		      if(outl[datacnt1].pkttype == 3) { /* Find first data   */
			datacnt = datacnt1;             /* before the ACK    */
			stopflag = 0;
		      } else {
			datacnt1--;
		      }
		    }
		    datacnt1--;
		    stopflag = 1;  /* Now be sure this was the right trigger */
		    while(stopflag && datacnt1 >= i) {
		      if((outl[datacnt1].pkttype == 3 && 
			  outl[datacnt1].seqn > outl[datacnt].seqn) || 
			 (outl[datacnt1].pkttype == 2 && 
			  outl[datacnt1].seqn + clientmss >= outl[datacnt].seqn)) {
			stopflag = 0;
		      } else {
			if(outl[datacnt1].pkttype == 3) datacnt = datacnt1;
			datacnt1--;
		      }
		    }
		    stopflag = 1;
		    lastdrop = datacnt - 1; /* Find location of drop        */
		    while(stopflag && lastdrop >= i) {
		      if(outl[lastdrop].pkttype == 3 &&
			 outl[datacnt].seqn > outl[lastdrop].seqn) {
			stopflag = 0;
		      } else {
			lastdrop--;
		      }
		    }
		    xfr[k].data = datacnt;
		    xfr[k].founddrop = 1;
		    xfr[k].drop = lastdrop;
		    if(xfr[k].phantom == 1) {
		      xfr[k].droptype = 4;
		    } else {
		      xfr[k].droptype = 5;
		    }
		  }
		}  /* End of phantom drop data assignment loop */

	      if(DEBUG==4)
		for(k=0;k<=rndindx;k++) {
		  if(xfr[k].founddrop) {
		    printf("Rnd %ld, ACK = %lu, Data = %lu Drop = %lu Type = %d PF= %d\n", 
			   k, outl[xfr[k].ack].seqn,outl[xfr[k].data].seqn, 
			   outl[xfr[k].drop].seqn, xfr[k].droptype, xfr[k].phantom);
		  } else {
		    printf("Rnd %ld, ACK = %lu, Data = %lu FD = %d Drop = 0\n", 
			   k, outl[xfr[k].ack].seqn,outl[xfr[k].data].seqn, 
			   xfr[k].founddrop);
		  }
		}

	      /* Now comes yet another pass through the data but this time  */
	      /* the idea is to go backwords through the rounds setting up  */
	      /* dependancies between rounds so that we can construct the   */
	      /* critical path.  The critical path through the rounds is    */
	      /* stored in the form of round indexes in the array rndnmbrs  */
	      /* NOTE:  skip round 0 since its already on the CP since its  */
	      /* actually part of the HTTP GET sequence...                  */

	      datacnt = 0;
	      for(k=0;k<MAXSEQ;k++) rndnmbrs[k] = 0; /*Init round array indx*/

	      /* First put any TO drops on CP which are not in data rounds  */
	      for(k=maxrnd;k>rndindx;k--) {
		if(xfr[k].droptype == 5) {
		  rndnmbrs[datacnt] = k;
		  if(DEBUG==4) 
		    printf("CP: Rnd %ld, Time out drop\n", k);
		  datacnt++;
		}
	      }
	      /* Now search back to round 1 (don't include round 0)         */
	      for(k=rndindx;k > 0;k--)
		{ /* Search for round in which last data pkt was triggered  */
		  if(outl[maxptr].seqn > xfr[k].bytestrt && 
		     outl[maxptr].seqn <= xfr[k].byteend) {
		    rndnmbrs[datacnt] = k;
		    if(DEBUG==4) 
		     printf("CP: Rnd %ld, Data = %lu Bstrt = %ld Bend = %ld NxtD = %ld\n",
			    k, outl[maxptr].seqn,xfr[k].bytestrt, xfr[k].byteend, 
			    outl[xfr[k].data].seqn);
		    maxptr = xfr[k].data;
		    datacnt++;
		  } else {
		    /*Put all time out drops on the critical path regardless*/
		    /*since TCP must wait to get an ACK for TMO retransmits */
		    /*Also put fast retrans drops on CP which are longer    */
		    /*than DROPdelay since phantom drops can result in clnt */
		    /*delays which are not actually there...                */
		    if(xfr[k].droptype == 5) {
		      rndnmbrs[datacnt] = k;
		      if(DEBUG==4) 
			printf("CP: Rnd %ld, Time out drop\n", k);
		      datacnt++;
		    }
		    if(xfr[k].droptype == 4 && 
		       outl[xfr[k].data].y2 - outl[xfr[k].drop].y2 >= DROPdelay) {
		      rndnmbrs[datacnt] = k;
		      if(DEBUG==4) 
			printf("CP: Rnd %ld, FstRtrn drop\n", k);
		      datacnt++;
		    }
		  }
		}
	      if(datacnt > 0 && rndnmbrs[datacnt -1] > 2) {
		printf("Backwards search for data triggers unsuccessful:  port %ld\n",
		       outl[i].rport);
		if(abortflag == 0) abortflag = 4;
	      }

	      /* Now we have the critical path through the bulk data xfer   */
	      /* so all that has to be done at this point is to go through  */
	      /* the rounds that are on the critical path and assign the    */
	      /* delays to client, server, network or drops.                */

	      for(k1=datacnt-1; k1 >= 0;k1--)
		{ 
		  k = rndnmbrs[k1];
		  if(DEBUG==4)
		    printf("Crit path:  Data = %lu ACK %ld Bytestrt = %ld Byteend = %ld\n",
			   outl[xfr[k].data].seqn, outl[xfr[k].ack].seqn, xfr[k].bytestrt, 
			   xfr[k].byteend);
		  /* First check to see if there was a DROP in rnd          */
		  if(xfr[k].founddrop && xfr[k].droptype > 0) {
		    /* Check to see if there was srvr delay before drop     */
		    if(outl[xfr[k].drop].y2 > lasty) { /* Data from server  */
		      cp[cur_cp].y1 = lasty;
		      cp[cur_cp].y2 = outl[xfr[k].drop].y2;
		      cp[cur_cp].x1 = cp_x; /*set x value for cp timeline   */
		      cp[cur_cp].dlytype = 2;  /* Assign delay to server    */
		      cp[cur_cp].status = 2;  /* Status = GET/Bulk transfer */
		      cp[cur_cp].host = outl[xfr[k].drop].host;
		      cp[cur_cp].rport = outl[xfr[k].drop].rport;
		      cp[cur_cp].file = num_trans;
		      if(DEBUG==4) 
			printf("Server drop delay: round %ld start %f end %f\n",
			       k,cp[cur_cp].y1,cp[cur_cp].y2);
		      if(outl[xfr[k].drop].y2 - lasty > SRVRABORTdelay && 
			 abortflag == 0) {
			  printf("CRITPATH:  SRVR abort transaction case 2 ");
			  printf("Rport %ld IPID %ld Data Seqn %lu\n", 
				 outl[xfr[k].drop].rport,outl[xfr[k].drop].ipid, 
				 outl[xfr[k].drop].seqn);
			abortflag = 2;  /* If delay is too long, abort    */
		      }
		      cur_cp++;
		    }
		    /* Now assign the drop delay between drop and rexmit    */
		    if( outl[xfr[k].data].y2 > lasty) {
		      cp[cur_cp].x1 = cp_x; /*set x value for cp timeline   */
		      cp[cur_cp].status = 2;  /* Status = GET/Bulk transfer */
		      cp[cur_cp].host = outl[xfr[k].drop].host;
		      cp[cur_cp].ipid = outl[xfr[k].drop].ipid;
		      cp[cur_cp].rport = outl[xfr[k].drop].rport;
		      cp[cur_cp].file = num_trans;
		      if(xfr[k].droptype == 5) {
			cp[cur_cp].dlytype = 5; /*set drop type to time out */
		      } else {
			cp[cur_cp].dlytype = 4; /*set drop type to fstretrn */
		      }
		      cp[cur_cp].y2 = outl[xfr[k].data].y2;
		      if(lasty <= outl[xfr[k].drop].y2) {
			cp[cur_cp].y1 = outl[xfr[k].drop].y2;
		      } else {
			cp[cur_cp].y1 = lasty;
		      }
		      lasty = cp[cur_cp].y2;
		      if(DEBUG==4) 
			printf("Rxdata drop delay: round %ld start %f end %f\n",
			       k,cp[cur_cp].y1,cp[cur_cp].y2);
		      cur_cp++;
		    }
		  } else { /* Now deal with the case with NO DROP in a round*/
		    /* Check if there was a server delay in sending the data*/
		    if(outl[xfr[k].data].y2 > lasty) {   /* Data from servr */
		      cp[cur_cp].y1 = lasty;
		      cp[cur_cp].y2 = outl[xfr[k].data].y2;
		      cp[cur_cp].x1 = cp_x; /* set x value for cp time line */
		      cp[cur_cp].dlytype = 2;    /* Assign delay to server  */
		      cp[cur_cp].status = 2;  /* Status = GET/Bulk transfer */
		      cp[cur_cp].host = outl[xfr[k].data].host;
		      cp[cur_cp].rport = outl[xfr[k].data].rport;
		      cp[cur_cp].file = num_trans;
		      if(DEBUG==4) 
			printf("Server data delay: round %ld start %f end %f\n",
			       k,cp[cur_cp].y1,cp[cur_cp].y2);
		      if(outl[xfr[k].data].y2 - lasty > SRVRABORTdelay && 
			 abortflag == 0) {
			printf("CRITPATH:  SRVR abort transaction case 2 ");
			printf("Rport %ld IPID %ld Data Seqn %lu\n", 
			       outl[xfr[k].data].rport,outl[xfr[k].data].ipid, 
			       outl[xfr[k].data].seqn);
			abortflag = 2;  /* If delay is too long, abort      */
		      }
                      cur_cp++;
		    } 
		  }
		  /* Now assign the network delay to the data packet or to  */
		  /* the retransmit packet if it was a fast retransmit drop */
		  if(lasty < outl[xfr[k].data].y1) {
		    cp[cur_cp].y2 = outl[xfr[k].data].y1;
		    cp[cur_cp].x1 = cp_x; /* set x value for cp time line   */
		    if(lasty <= outl[xfr[k].data].y2) {
		      cp[cur_cp].y1 = outl[xfr[k].data].y2;
		    } else {
		      cp[cur_cp].y1 = lasty;
		    }
		    cp[cur_cp].dlytype = 1;  /* assign the network delay    */
		    if(DEBUG==4) 
		      printf("Net data delay: round %ld start %f end %f\n",
			     k,cp[cur_cp].y1,cp[cur_cp].y2);
		    cp[cur_cp].status = 2;  /* Status = GET/Bulk transfer   */
		    cp[cur_cp].host = outl[xfr[k].data].host;
		    cp[cur_cp].ipid = outl[xfr[k].data].ipid;
		    cp[cur_cp].rport = outl[xfr[k].data].rport;
		    lasty = cp[cur_cp].y2;
		    cp[cur_cp].file = num_trans;
		    cur_cp++;
		  }
		  /* Now deal with the ACK packet in the round              */
		  if(outl[xfr[k].ack].y1 > lasty) {
		    cp[cur_cp].y1 = lasty;
		    cp[cur_cp].y2 = outl[xfr[k].ack].y1;
		    cp[cur_cp].x1 = cp_x;    /* set x value for cp time line*/
		    cp[cur_cp].dlytype = 3; /* Assign delay to client       */
		    cp[cur_cp].status = 2;  /* Status = GET/Bulk transfer   */
		    cp[cur_cp].host = outl[xfr[k].ack].host;
		    cp[cur_cp].rport = outl[xfr[k].ack].rport;
		    cp[cur_cp].file = num_trans;
		    cp[cur_cp].src = cli_host;
		    if(DEBUG==4) 
		      printf("Client ack delay: round %ld start %f end %f\n",
			     k,cp[cur_cp].y1,cp[cur_cp].y2);
		    if(outl[xfr[k].ack].y1 - lasty > CLNTABORTdelay &&
		       abortflag == 0) {
		      printf("CRITPATH:  CLNT abort transaction case 2 ");
		      printf("Rport %ld IPID %ld Data Seqn %lu\n", 
			     outl[xfr[k].ack].rport,outl[xfr[k].ack].ipid, 
			     outl[xfr[k].ack].seqn);
		      abortflag = 1;  /* If delay is too long, abort    */
		    }
		    cur_cp++;
		  }
		  if(lasty < outl[xfr[k].ack].y2) {
		    cp[cur_cp].y2 = outl[xfr[k].ack].y2;
		    cp[cur_cp].x1 = cp_x;    /* set x value for cp time line*/
		    cp[cur_cp].dlytype = 1;  /* assign the network delay    */
		    cp[cur_cp].status = 2;  /* Status = GET/Bulk transfer   */
		    cp[cur_cp].host = outl[xfr[k].ack].host;
		    cp[cur_cp].ipid = outl[xfr[k].ack].ipid;
		    cp[cur_cp].rport = outl[xfr[k].ack].rport;
		    cp[cur_cp].file = num_trans;
		    cp[cur_cp].src = cli_host;
		    if(lasty <= outl[xfr[k].ack].y1) {
		      cp[cur_cp].y1 = outl[xfr[k].ack].y1;
		    } else {
		      cp[cur_cp].y1 = lasty;
		    }
		    lasty = cp[cur_cp].y2;
		    if(DEBUG==4) 
		      printf("Net ack delay: round %ld start %f end %f\n",
			     k,cp[cur_cp].y1,cp[cur_cp].y2);
		    cur_cp++;
		  }
		}

	      /* Finally, get things ready for connection tear down by      */
	      /* moving the index up to where the first FIN is sent...      */

	      founddrop = 0;  /* Just in case founddrop was set to 1        */
	      stopflag = 1;
	      j = i;
	      while(stopflag && j < tlend)
		if(outl[j].pkttype == 4 || outl[j].origpkt == 4) {
		  stopflag = 0;
		} else {
		  j++;
		}
	      i = j - 1;
	      /* At this point we must again make sure that the packet     */
	      /* order is FIN1, ACK1, FIN2, ACK2.  Since packets are       */
	      /* sorted by server timestamp at this point, the ACK1,       */
	      /* FIN2 and ACK2 can all have the same time - so make sure   */
	      /* they are in the right order...                            */
		if(outl[j+1].y1 == outl[j+2].y1 && 
		   outl[j+1].y1 == outl[j+3].y1) /*need to check next 3 pkt*/
		  { 
		    j = i + 2;
		    while(outl[i + 2].y1 == outl[j + 1].y1 && j < tlend) j++;
		    qsort(&outl[i+2],j - i + 1,sizeof(outl[0]),compare6);
		    j = i + 2;
		    while((outl[i + 2].y1 == outl[j + 1].y1) &&
			  (outl[i + 2].host == outl[j + 1].host) &&
			  j < tlend) j++;
		    if(j - i + 1>= 2)  /*Be sure seq=> ACK1 FIN2 from clnt */
		      qsort(&outl[i + 2],j - i + 1,sizeof(outl[0]),compare3);
		  }
	      httpstate++;  
	      break;

	    case 3:  /* TCP connection tear down sequence - 4 packets       */
	      /* This routine can be tricky since the FIN from server, ACK  */
	      /* from client and FIN from client can all have same time     */
	      /* stamp at client. Thus we have to do a sort initial pkts    */
	      /* There can also be a problem of ACK's from the data transfer*/
	      /* showing up in the FIN sequence - these must be filtered out*/

	      /*If pkt is a drop find retransmt but only if it was not an ACK*/
	      if(outl[i].pkttype == 6 && outl[i].origpkt != 2) { 
		founddrop = 1;
		dropseqn[0] = outl[i].seqn;
		cp[cur_cp].y1 = outl[i].y1;
		cp[cur_cp].dlytype = 4;  /* Assign delay to type drop       */
		cp[cur_cp].status = 3;  /* Set status connection tear down  */
		cp[cur_cp].host = outl[i].host;
		cp[cur_cp].ipid = outl[i].ipid;
		cp[cur_cp].rport = outl[i].rport;
		cp[cur_cp].file = num_trans;
		if(outl[i].host == 1) {  /* Assign host type to cp          */
		  cp[cur_cp].src = cli_host;
		}
	      } else {
		if((outl[i].pkttype == 4 || outl[i].origpkt == 4) 
		   && outl[i].host == 0) 
		  {            /* This is the first FIN from the server     */
		    if(finstate == 0) { /*If finstate = 1 its the first FIN */
		      finackseq = outl[i].seqn;
		      if(outl[i].y2 > lasty) {
			cp[cur_cp].y1 = lasty;
			cp[cur_cp].y2 = outl[i].y2;
			cp[cur_cp].dlytype = 2;  /* Assign delay to server  */
			cp[cur_cp].status = 3;  /* Status connect teardown  */
			cp[cur_cp].host = outl[i].host;
			cp[cur_cp].rport = outl[i].rport;
			cp[cur_cp].file = num_trans;
			if(outl[i].y2 - lasty > SRVRABORTdelay && abortflag ==0) {
			  printf("CRITPATH:  SRVR abort transaction case 3 ");
			  printf("Rport %ld IPID %ld Data Seqn %lu\n", 
				 outl[i].rport,outl[i].ipid, outl[i].seqn);
			  abortflag = 2;  /* If delay is too long, abort    */
			}
			cur_cp++;
		      }
		      cp[cur_cp].y1 = outl[i].y2;
		      cp[cur_cp].y2 = outl[i].y1;
		      finstate++;
		    } else {    /* If finstate > 1 its a timeout/retransmit */
		      cp[cur_cp].y1 = lasty;
		      cp[cur_cp].dlytype = 5; /* Assign delay to retransmit */
		      cp[cur_cp].status = 3;  /* Set status connect teardown*/
		      cp[cur_cp].host = outl[i].host;
		      cp[cur_cp].ipid = outl[i].ipid;
		      cp[cur_cp].rport = outl[i].rport;  
		      cp[cur_cp].y2 = outl[i].y2;
		      cur_cp++;
		      cp[cur_cp].y1 = outl[i].y2;
		      cp[cur_cp].y2 = outl[i].y1;
		    }
		    cp[cur_cp].x1 = cp_x;  /* set x value for cp time line  */
		    cp[cur_cp].dlytype = 1;  /* assign the network delay    */
		    cp[cur_cp].status = 3;  /* Set status connect teardown  */
		    cp[cur_cp].host = outl[i].host;
		    cp[cur_cp].ipid = outl[i].ipid;
		    cp[cur_cp].rport = outl[i].rport;
		    cp[cur_cp].file = num_trans;
		    lasty = cp[cur_cp].y2;
		    cur_cp++;
		  }
		if(outl[i].pkttype == 2 && outl[i].host == 1 && 
		   (outl[i].seqn == finackseq + 1 || numdupack > 0)) 
		  {          /*  This is the ACK of the FIN from the client */
		    /* Actually, this FIN-ACK SHOULD NEVER be on the CP - PB*/
		    /***************************************************
		    if(outl[i].y1 > lasty) {
		      cp[cur_cp].y1 = lasty;
		      cp[cur_cp].y2 = outl[i].y1;
		      cp[cur_cp].dlytype = 3;
		      cp[cur_cp].status = 3;
		      cp[cur_cp].host = outl[i].host;
		      cp[cur_cp].rport = outl[i].rport;
		      cp[cur_cp].src = cli_host;
		      cp[cur_cp].file = num_trans;
		      if(outl[i].y1 - lasty > CLNTABORTdelay && abortflag == 0) {
			printf("CRITPATH:  CLNT abort transaction case 3 ");
			printf("Rport %ld IPID %ld Data Seqn %lu\n", 
			       outl[i].rport,outl[i].ipid, outl[i].seqn);
			abortflag = 1;
		      }
		      cur_cp++;
		    }
		    cp[cur_cp].y1 = outl[i].y1;
		    cp[cur_cp].y2 = outl[i].y2;
		    cp[cur_cp].x1 = cp_x;
		    cp[cur_cp].dlytype = 1;
		    cp[cur_cp].status = 3;
		    cp[cur_cp].host = outl[i].host;
		    cp[cur_cp].ipid = outl[i].ipid;
		    cp[cur_cp].rport = outl[i].rport;
		    cp[cur_cp].src = cli_host;
		    lasty = cp[cur_cp].y2;
		    cp[cur_cp].file = num_trans;
		    cur_cp++;
		    *****************************************************/
		    finstate++;
		  }
		if(outl[i].pkttype == 4 && outl[i].host == 1) 
		  {                  /* This is the FIN from the client     */
		    finackseq = outl[i].seqn;
		    if(finstate == 2) { /*If finstate = 2 its the first FIN */
		      if(outl[i].y1 > lasty) {
			cp[cur_cp].y1 = lasty;
			cp[cur_cp].y2 = outl[i].y1;
			cp[cur_cp].dlytype = 3;  /* Assign delay to client  */
			cp[cur_cp].status = 3;  /* Status connect teardown  */
			cp[cur_cp].host = outl[i].host;
			cp[cur_cp].rport = outl[i].rport;
			cp[cur_cp].src = cli_host;
			cp[cur_cp].file = num_trans;
			if(outl[i].y1 - lasty > CLNTABORTdelay && abortflag ==0) {
			  printf("CRITPATH:  CLNT abort transaction case 3 ");
			  printf("Rport %ld IPID %ld Data Seqn %lu\n", 
				 outl[i].rport,outl[i].ipid, outl[i].seqn);
			  abortflag = 1;  /* If delay is too long, abort    */
			}
			cur_cp++;
		      }
		      cp[cur_cp].y1 = outl[i].y1;
		      cp[cur_cp].y2 = outl[i].y2;
		      finstate++;
		    } else {    /* If finstate > 2 its a timeout/retransmit */
		      cp[cur_cp].y1 = lasty;
		      cp[cur_cp].dlytype = 5; /* Assign delay to retransmit */
		      cp[cur_cp].status = 3;  /* Set status connect teardown*/
		      cp[cur_cp].host = outl[i].host;
		      cp[cur_cp].ipid = outl[i].ipid;
		      cp[cur_cp].rport = outl[i].rport;  
		      cp[cur_cp].y2 = outl[i].y1;
		      cp[cur_cp].src = cli_host;
		      cp[cur_cp].file = num_trans;
		      cur_cp++;
		      cp[cur_cp].y1 = outl[i].y1;
		      cp[cur_cp].y2 = outl[i].y2;
		    }
		    cp[cur_cp].x1 = cp_x;  /* set x value for cp time line*/
		    cp[cur_cp].dlytype = 1;  /* Assign the network delay  */
		    cp[cur_cp].status = 3;  /* Set status connect teardown*/
		    cp[cur_cp].host = outl[i].host;
		    cp[cur_cp].ipid = outl[i].ipid;
		    cp[cur_cp].rport = outl[i].rport;
		    cp[cur_cp].src = cli_host;
		    cp[cur_cp].file = num_trans;
		    lasty = cp[cur_cp].y2;
		    cur_cp++;
		  }
		if(outl[i].pkttype == 2 && outl[i].host == 0 &&
		   outl[i].seqn == finackseq + 1) 
		  {     /*  This is the final FIN - ACK from the server     */
		    if(outl[i].y2 > lasty) {
		      cp[cur_cp].y1 = lasty;
		      cp[cur_cp].y2 = outl[i].y2;
		      cp[cur_cp].dlytype = 2;   /* Assign delay to server   */
		      cp[cur_cp].status = 3;  /* Set status connect teardown*/
		      cp[cur_cp].host = outl[i].host;
		      cp[cur_cp].rport = outl[i].rport;
		      cp[cur_cp].file = num_trans;
		      if(outl[i].y2 - lasty > SRVRABORTdelay && abortflag == 0) {
			printf("CRITPATH:  SRVR abort transaction case 3 ");
			printf("Rport %ld IPID %ld Data Seqn %lu\n", 
			       outl[i].rport,outl[i].ipid, outl[i].seqn);
			abortflag = 2;  /* If delay is too long, abort    */
		      }
		      cur_cp++;
		    }
		    cp[cur_cp].y1 = outl[i].y2;
		    cp[cur_cp].y2 = outl[i].y1;
		    cp[cur_cp].x1 = cp_x;  /* set x value for cp time line  */
		    cp[cur_cp].dlytype = 1;  /* Assign the network delay    */
		    cp[cur_cp].status = 3;  /* Set status connect teardown  */
		    cp[cur_cp].host = outl[i].host;
		    cp[cur_cp].ipid = outl[i].ipid;
		    cp[cur_cp].rport = outl[i].rport;
		    cp[cur_cp].file = num_trans;
		    lasty = cp[cur_cp].y2;
		    finstate++;
		    cur_cp++;
		    httpstate++;
		  }
	      }
	      break;
	    default:
	      if(httpstate == 4 && outl[i].pkttype != 7) {
		printf("Extra packet found in rport %ld after final ack %d\n",
		       outl[i].rport, outl[i].pkttype);
	      } else {
		printf("Unknown HTTP state in function critpath - aborting\n");
		exit(0);
	      }
	      break;
	    }
	} else {
	  /* A drop has been found so loop until retransmit is found       */
	  /* This loop is only used for a drop during connect setup, http  */
	  /* get or connect tear down - any drop during those periods is   */
	  /* a time out drop.                                              */

	  /* There is a weird ordering problem which can exist during conct*/
	  /* set up.  The ordering must be SYN, SYN-ACK, ACK.  If the SYN- */
	  /* ACK is dropped, be sure that order goes SYN-ACK, ACK so sort  */
	  /* by packet type...                                             */

	  if(httpstate == 0 && synstate == 1) {
	    j = i + 1;
	    if(outl[i + 1].y1 == outl[j + 1].y1) {
	      while(outl[i + 1].y1 == outl[j + 1].y1 && j < tlend) j++;
	      qsort(&outl[i + 1],j - i,sizeof(outl[0]),compare3);
	    }
	  }

	  if(dropseqn[0] == outl[i].seqn && outl[i].pkttype != 6)
	    {
	      if(outl[i].y1 > outl[i].y2) { /* Set end time for drop period*/
		cp[cur_cp].y2 = outl[i].y2;
		lasty = outl[i].y2;
	      } else {
		cp[cur_cp].y2 = outl[i].y1;
		lasty = outl[i].y1;
	      }
	      cp[cur_cp].dlytype = 5;         /* delay type and set cwnd   */
	      cp[cur_cp].file = num_trans;
	      cur_cp++;
	      founddrop = 0;
	      i--;    /* Be sure to capture data for retransmitted packet  */ 
	    }
	}
      i++;  /* Increment counter index */
    }
  if(abortflag > 0) { /* Set the abort flag for the current transfer       */
    for(i=cp_start;i<cur_cp;i++)
      cp[i].abort = abortflag;  /* Abort > 1 means xfer will not be in stats*/
  }

/* End function critpath */
}
